# coding: utf-8

import unittest

from nxtomomill.io import utils
from nxtomomill.io.framegroup import FrameGroup


class TestConvertStrToTuple(unittest.TestCase):
    """
    test convert_str_to_tuple function
    """

    def testStr1(self):
        self.assertEqual(utils.convert_str_to_tuple("toto, tata"), ("toto", "tata"))

    def testStr2(self):
        self.assertEqual(
            utils.convert_str_to_tuple("'toto', \"tata\""), ("toto", "tata")
        )

    def testStr3(self):
        self.assertEqual(utils.convert_str_to_tuple("test"), ("test",))

    def testStr4(self):
        self.assertEqual(
            utils.convert_str_to_tuple("(this is a test)"), ("this is a test",)
        )

    def testStr5(self):
        self.assertEqual(
            utils.convert_str_to_tuple("(this is a test, 'and another one')"),
            ("this is a test", "and another one"),
        )


class TestIsUrlPath(unittest.TestCase):
    """test the is_url function"""

    def test_invalid_url_1(self):
        self.assertFalse(utils.is_url_path("/toto/tata"))

    def test_invalid_url_2(self):
        self.assertFalse(utils.is_url_path("tata"))

    def test_valid_url_1(self):
        self.assertTrue(utils.is_url_path("silx:///data/image.h5?path=/scan_0/data"))

    def test_valid_url_2(self):
        self.assertTrue(utils.is_url_path("silx:///data/image.edf"))

    def test_valid_url_3(self):
        self.assertTrue(utils.is_url_path("silx://image.h5"))

    def test_valid_url_4(self):
        self.assertTrue(
            utils.is_url_path("silx:///data/image.h5?path=/scan_0/data&slice=1,5")
        )


class TestConvertStrToFrameGrp(unittest.TestCase):
    """
    test the `convert_str_to_frame_grp` function
    """

    def test_valid_1(self):
        dg1 = FrameGroup(
            frame_type="dark", url="silx:///file.h5?data_path=/dark", copy=True
        )
        dg2 = FrameGroup(
            frame_type="flat", url="silx:///file.h5?data_path=/flat", copy=True
        )
        dg3 = FrameGroup(
            frame_type="projection",
            url="silx:///file.h5?data_path=/projection1",
            copy=True,
        )
        dg4 = FrameGroup(
            frame_type="alignment",
            url="silx:///file.h5?data_path=/projection2",
            copy=True,
        )

        my_str = "({dg1}, {dg2}, {dg3}, {dg4}, )".format(
            dg1=dg1, dg2=dg2, dg3=dg3, dg4=dg4
        )
        frame_grps = utils.convert_str_to_frame_grp(my_str)
        self.assertTrue(isinstance(frame_grps, tuple))
        self.assertTrue(len(frame_grps) == 4)
        for o_fg, i_fg in zip((dg1, dg2, dg3, dg4), frame_grps):
            self.assertEqual(str(o_fg), str(i_fg))

    def test_valid_2(self):
        pass

    def test_invalid_1(self):
        pass



# coding: utf-8

import os
import shutil
import tempfile
import unittest

from nxtomomill.io.config.confighandler import TomoHDF5ConfigHandler
from nxtomomill.io.config.hdf5config import TomoHDF5Config
from nxtomomill.io.framegroup import FrameGroup


class _ARParseMock(object):
    def __init__(self):
        self.config = None
        self.input_file = None
        self.output_file = None
        self.set_params = None
        self.align_titles = None
        self.proj_titles = None
        self.flat_titles = None
        self.dark_titles = None
        self.init_zserie_titles = None
        self.init_multi_tomo_titles = None
        self.init_titles = None
        self.init_back_and_forth_titles = None
        self.x_pixel_size_key = None
        self.y_pixel_size_key = None
        self.acq_expo_time_keys = None
        self.rot_angle_keys = None
        self.valid_camera_names = None
        self.translation_z_keys = None
        self.sample_y_keys = None
        self.sample_x_keys = None
        self.request_input = False
        self.raises_error = False
        self.duplicate_data = None
        self.ignore_sub_entries = None
        self.entries = None
        self.debug = False
        self.overwrite = False
        self.single_file = False
        self.no_master_file = False
        self.file_extension = None
        self.field_of_view = None
        self.sample_detector_distance_paths = None


class TestH5Config(unittest.TestCase):
    """
    Test the HDF5Config class
    """

    def setUp(self) -> None:
        self.folder = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)

    def test_creation_from_config_file(self):
        options = _ARParseMock()
        with self.assertRaises(ValueError):
            TomoHDF5ConfigHandler(argparse_options=options)

        input_file_path = os.path.join(self.folder, "output_file.cfg")
        input_config = TomoHDF5Config()
        input_config.to_cfg_file(input_file_path)

        options.config = input_file_path

        with self.assertRaises(ValueError):
            TomoHDF5ConfigHandler(argparse_options=options)

        # insure this is valid if we add an input file and an output file
        # from the command line
        options.input_file = "toto.h5"
        options.output_file = "toto.nx"
        TomoHDF5ConfigHandler(argparse_options=options)

        # try providing twice an input file
        input_config.input_file = "toto2.h5"
        input_config.to_cfg_file(input_file_path)

        with self.assertRaises(ValueError):
            TomoHDF5ConfigHandler(argparse_options=options)

        # or if we provide them from the configuration file
        options.input_file = None
        options.output_file = None
        input_config.input_file = "toto.h5"
        input_config.output_file = "toto.nx"
        input_config.to_cfg_file(input_file_path)
        TomoHDF5ConfigHandler(argparse_options=options)

        # try providing twice an output file
        input_config.input_file = None
        input_config.output_file = "toto2.nx"
        input_config.to_cfg_file(input_file_path)
        with self.assertRaises(ValueError):
            TomoHDF5ConfigHandler(argparse_options=options)

        # try another random para,eter
        input_config.output_file = None
        input_config.sample_x_keys = ("xtrans2",)
        options.sample_x_keys = "xtrans3"
        input_config.to_cfg_file(input_file_path)
        with self.assertRaises(ValueError):
            TomoHDF5ConfigHandler(argparse_options=options)

    def test_creation_from_cmd_line_opts(self):
        options = _ARParseMock()
        input_file_path = os.path.join(self.folder, "output_file.cfg")
        input_config = TomoHDF5Config()
        input_config.to_cfg_file(input_file_path)
        options.input_file = "toto.h5"
        options.output_file = "toto.nx"
        options.debug = True
        TomoHDF5ConfigHandler(argparse_options=options)

        options.config = input_config
        with self.assertRaises(ValueError):
            TomoHDF5ConfigHandler(argparse_options=options)


class TestFrameUrls(unittest.TestCase):
    """Test frames urls"""

    def setUp(self) -> None:
        self.folder = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)

    def write_and_read_configuration(self, data_urls):
        configuration = TomoHDF5Config().to_dict()
        configuration[TomoHDF5Config.FRAME_TYPE_SECTION_DK][TomoHDF5Config.DATA_DK] = (
            FrameGroup.list_to_str(data_urls)
        )
        file_path = os.path.join(self.folder, "h52nx.cfg")
        TomoHDF5Config.dict_to_cfg(file_path=file_path, dict_=configuration)

        options = _ARParseMock()
        options.input_file = "toto.h5"
        options.output_file = "toto.nx"
        options.config = file_path
        config_handler = TomoHDF5ConfigHandler(options)
        return config_handler.configuration

    def test_local_paths(self):
        # TODO: test all option to read / write the configuration
        res_config = self.write_and_read_configuration(
            (
                FrameGroup(url="/entry0000/data/flats", frame_type="flat"),
                FrameGroup(url="/entry0000/data/projection1", frame_type="projection"),
                FrameGroup(url="/entry0000/data/projection2", frame_type="projection"),
                FrameGroup(url="/entry0000/data/alignment1", frame_type="alignment"),
                FrameGroup(url="/entry0000/data/alignment2", frame_type="alignment"),
                FrameGroup(url="/entry0000/data/darks", frame_type="dark"),
            )
        )
        self.assertNotEqual(res_config, None)
        self.assertEqual(len(res_config.data_frame_grps), 6)
        # check flats
        flat_url = res_config.data_frame_grps[0].url
        self.assertTrue(flat_url.file_path() in (None, ""))
        self.assertEqual(flat_url.data_path(), "/entry0000/data/flats")
        self.assertTrue(flat_url.data_slice() in (None, ""))
        # check projections
        projection_url_0 = res_config.data_frame_grps[1].url
        projection_url_1 = res_config.data_frame_grps[2].url
        self.assertTrue(projection_url_0.file_path() in (None, ""))
        self.assertEqual(projection_url_0.data_path(), "/entry0000/data/projection1")
        self.assertEqual(projection_url_0.data_slice(), None)
        self.assertTrue(projection_url_1.file_path() in (None, ""))
        self.assertEqual(projection_url_1.data_path(), "/entry0000/data/projection2")
        self.assertEqual(projection_url_1.data_slice(), None)

        # check darks
        dark_url = res_config.data_frame_grps[-1].url
        self.assertTrue(dark_url.file_path() in (None, ""))
        self.assertEqual(dark_url.data_path(), "/entry0000/data/darks")
        self.assertEqual(dark_url.data_slice(), None)

        # check alignments
        alignment_url_0 = res_config.data_frame_grps[3].url
        alignment_url_1 = res_config.data_frame_grps[4].url
        self.assertTrue(alignment_url_0.file_path() in (None, ""))
        self.assertEqual(alignment_url_0.data_path(), "/entry0000/data/alignment1")
        self.assertEqual(alignment_url_0.data_slice(), None)
        self.assertEqual(alignment_url_1.data_path(), "/entry0000/data/alignment2")

    def test_external_paths(self):
        """
        Check that DataUrl are handled.

        Warning: this also check data_slices but those are not handled.
        """
        res_config = self.write_and_read_configuration(
            (
                FrameGroup(
                    frame_type="flat",
                    url="silx:///myfile.h5?path=/entry0000/data/flats",
                ),
                FrameGroup(
                    frame_type="dark",
                    url="h5py:///data/file2.hdf5?path=/entry0000/data/darks",
                ),
                FrameGroup(frame_type="proj", url="h5py:///data/file3.hdf5?path=/data"),
                FrameGroup(
                    frame_type="proj",
                    url="silx:///data/file2.hdf5?path=/entry0000/data/projection2&slice=5:100",
                ),
                FrameGroup(
                    frame_type="alignment",
                    url="silx:///myfile.h5?path=/entry0000/data/alignment&slice=5",
                ),
            )
        )
        self.assertNotEqual(res_config, None)
        self.assertEqual(len(res_config.data_frame_grps), 5)
        # check flats
        flat_url = res_config.data_frame_grps[0]
        self.assertEqual(flat_url.url.file_path(), "/myfile.h5")
        self.assertEqual(flat_url.url.data_path(), "/entry0000/data/flats")
        self.assertEqual(flat_url.url.data_slice(), None)
        self.assertEqual(flat_url.url.scheme(), "silx")

        # check projections
        projection_url_0 = res_config.data_frame_grps[2].url
        projection_url_1 = res_config.data_frame_grps[3].url
        self.assertEqual(projection_url_0.file_path(), "/data/file3.hdf5")
        self.assertEqual(projection_url_0.data_path(), "/data")
        self.assertEqual(projection_url_0.data_slice(), None)
        self.assertEqual(projection_url_0.scheme(), "h5py")
        self.assertEqual(projection_url_1.file_path(), "/data/file2.hdf5")
        self.assertEqual(projection_url_1.data_path(), "/entry0000/data/projection2")
        self.assertEqual(projection_url_1.data_slice(), (slice(5, 100),))
        self.assertEqual(projection_url_1.scheme(), "silx")

        # check darks
        dark_url = res_config.data_frame_grps[1].url
        self.assertEqual(dark_url.file_path(), "/data/file2.hdf5")
        self.assertEqual(dark_url.data_path(), "/entry0000/data/darks")
        self.assertEqual(dark_url.data_slice(), None)
        self.assertEqual(dark_url.scheme(), "h5py")

        # check alignments
        alignment_url_0 = res_config.data_frame_grps[4].url
        self.assertEqual(alignment_url_0.file_path(), "/myfile.h5")
        self.assertEqual(alignment_url_0.data_path(), "/entry0000/data/alignment")
        self.assertEqual(alignment_url_0.data_slice(), (5,))
        self.assertEqual(alignment_url_0.scheme(), "silx")



from __future__ import annotations

import h5py
from silx.io.utils import h5py_read_dataset

__all__ = [
    "TomoConfig",
]


class TomoConfig:
    """
    hold motor used for tomography acquisition according to https://tomo.gitlab-pages.esrf.fr/bliss-tomo/master/modelization_sample_stage.html convension
    """

    def __init__(self) -> None:
        self._rotation = None
        self._sample_u = None
        self._sample_v = None
        self._sample_x = None
        self._sample_y = None
        self._translation_x = None
        self._translation_y = None
        self._translation_z = None
        self._tomo_detector = None

    @property
    def rotation(self) -> str | None:
        return self._rotation

    @rotation.setter
    def rotation(self, motor: str | None):
        self._rotation = motor

    @property
    def sample_u(self) -> str | None:
        return self._sample_u

    @sample_u.setter
    def sample_u(self, motor: str | None):
        self._sample_u = motor

    @property
    def sample_v(self) -> str | None:
        return self._sample_v

    @sample_v.setter
    def sample_v(self, motor: str | None):
        self._sample_v = motor

    @property
    def sample_x(self) -> str | None:
        return self._sample_x

    @sample_x.setter
    def sample_x(self, motor: str | None):
        self._sample_x = motor

    @property
    def sample_y(self) -> str | None:
        return self._sample_y

    @sample_y.setter
    def sample_y(self, motor: str | None):
        self._sample_y = motor

    @property
    def translation_x(self) -> str | None:
        return self._translation_x

    @translation_x.setter
    def translation_x(self, motor: str | None):
        self._translation_x = motor

    @property
    def translation_y(self) -> str | None:
        return self._translation_y

    @translation_y.setter
    def translation_y(self, motor: str | None):
        self._translation_y = motor

    @property
    def translation_z(self) -> str | None:
        return self._translation_z

    @translation_z.setter
    def translation_z(self, motor: str | None):
        self._translation_z = motor

    @property
    def tomo_detector(self) -> str | None:
        return self._tomo_detector

    @tomo_detector.setter
    def tomo_detector(self, detector_name):
        self._tomo_detector = detector_name

    def __str__(self) -> str:
        return "tomo_config:" + " ; ".join(
            [
                f"rotation={self.rotation}",
                f"sample_u={self.sample_u}",
                f"sample_v={self.sample_v}",
                f"sample_x={self.sample_x}",
                f"sample_y={self.sample_y}",
                f"tomo_detector={self.tomo_detector}",
                f"translation_x={self.translation_x}",
                f"translation_y={self.translation_y}",
                f"translation_z={self.translation_z}",
            ]
        )

    @staticmethod
    def from_technique_group(technique_group: h5py.Group):
        """
        get rotation motor and thinks like this from the 'tomoconfig'.
        This can retrieve one or several dataset name or a single one.
        In the case of several dataset name we get (real_motor_name, bliss_alias)

        If the motor moves then this is pretty simple the real motor_name dataset exists.
        But if the motors does not move during the bliss scan (scalar value) then the real_motor_name dataset doesn't exists and
        the bliss alias does. This is why we need to keep both and check both during the 'standard process'...
        """
        if not isinstance(technique_group, h5py.Group):
            raise TypeError(
                f"instrument_group is expected to be an instance of {h5py.Group}. {type(technique_group)} provided"
            )
        if "tomoconfig" not in technique_group:
            raise KeyError("could find 'tomoconfig' key")
        else:
            tomo_config_group = technique_group.get("tomoconfig")

        def get_dataset(group, dataset_name, default):
            if dataset_name not in group:
                return default
            else:
                return h5py_read_dataset(group[dataset_name])

        tomo_config = TomoConfig()
        tomo_config.rotation = get_dataset(tomo_config_group, "rotation", None)
        tomo_config.sample_u = get_dataset(tomo_config_group, "sample_u", None)
        tomo_config.sample_v = get_dataset(tomo_config_group, "sample_v", None)
        tomo_config.sample_x = get_dataset(tomo_config_group, "sample_x", None)
        tomo_config.sample_y = get_dataset(tomo_config_group, "sample_y", None)
        tomo_config.tomo_detector = get_dataset(tomo_config_group, "detector", None)
        # translation x == sample_u
        tomo_config.translation_x = get_dataset(
            tomo_config_group, "sample_u", None
        )  # Needs translation X mapping
        tomo_config.translation_y = get_dataset(
            tomo_config_group, "translation_y", None
        )
        tomo_config.translation_z = get_dataset(
            tomo_config_group, "translation_z", None
        )

        return tomo_config



# coding: utf-8

import os

import pytest

from nxtomo.nxobject.nxdetector import FieldOfView
from nxtomo.nxobject.nxsource import ProbeType, SourceType
from nxtomomill.io.config.fluoconfig import TomoFluoConfig, generate_default_fluo_config


def test_TomoFluoConfig_to_dict(tmp_path):
    """Test processing of the to_dict"""
    config = TomoFluoConfig()

    config_dict = config.to_dict()
    for key in (
        TomoFluoConfig.GENERAL_SECTION_DK,
        TomoFluoConfig.SOURCE_SECTION_DK,
    ):
        assert key in config_dict

    TomoFluoConfig.from_dict(config_dict)

    file_path = os.path.join(tmp_path, "config.cfg")
    assert not os.path.exists(file_path)
    config.to_cfg_file(file_path)
    assert os.path.exists(file_path)
    config_loaded = TomoFluoConfig.from_cfg_file(file_path=file_path)

    assert config_loaded.to_dict() == config.to_dict()


def test_TomoFluoConfig_default_config(tmp_path):
    """insure default configuration generation works"""
    file_path = os.path.join(tmp_path, "config.cfg")
    assert not os.path.exists(file_path)
    config = generate_default_fluo_config()
    assert isinstance(config, dict)
    TomoFluoConfig.from_dict(config).to_cfg_file(file_path=file_path)
    assert os.path.exists(file_path)


def test_TomoFluoConfig_setters():
    """test the different setters and getter of the EDFTomoConfig"""
    config = TomoFluoConfig()

    # try to test a new attribut (insure class is frozeen)
    with pytest.raises(AttributeError):
        config.new_attrs = "toto"

    # test general section setters
    with pytest.raises(TypeError):
        config.input_folder = 12.0
    config.input_folder = "my_folder"
    with pytest.raises(TypeError):
        config.output_file = 12.0
    config.output_file = "my_nx.nx"

    config.file_extension = ".nx"

    with pytest.raises(TypeError):
        config.detectors = 12.0
    config.detectors = None
    config.detectors = "test_"

    with pytest.raises(TypeError):
        config.dataset_basename = 12.0
    config.dataset_basename = None
    config.dataset_basename = "test_"

    config.overwrite = True

    with pytest.raises(TypeError):
        config.title = 12.0
    config.title = None
    config.title = "my title"

    # test source setters
    config.instrument_name = None
    config.instrument_name = "BMXX"
    with pytest.raises(TypeError):
        config.instrument_name = 12.3

    config.source_name = None
    config.source_name = "ESRF"
    with pytest.raises(TypeError):
        config.source_name = 12.2
    config.source_name = "XFEL"

    config.source_type = None
    config.source_type = SourceType.FIXED_TUBE_X_RAY.value
    config.source_type = SourceType.FIXED_TUBE_X_RAY
    with pytest.raises(ValueError):
        config.source_type = "trsts"
    with pytest.raises(TypeError):
        config.source_type = 123

    config.source_probe = None
    config.source_probe = ProbeType.ELECTRON
    config.source_probe = "positron"
    with pytest.raises(TypeError):
        config.source_probe = 123

    # test detector setters
    config.field_of_view = None
    with pytest.raises(TypeError):
        config.field_of_view = 12
    with pytest.raises(ValueError):
        config.field_of_view = "toto"
    config.field_of_view = FieldOfView.FULL.value
    config.field_of_view = FieldOfView.HALF
    # TO DO : add test on detector setter

    config_dict = config.to_dict()
    general_section_dict = config_dict[TomoFluoConfig.GENERAL_SECTION_DK]
    assert general_section_dict[TomoFluoConfig.INPUT_FOLDER_DK] == "my_folder"
    assert general_section_dict[TomoFluoConfig.OUTPUT_FILE_DK] == "my_nx.nx"
    assert general_section_dict[TomoFluoConfig.FILE_EXTENSION_DK] == ".nx"
    assert general_section_dict[TomoFluoConfig.OVERWRITE_DK] is True
    assert general_section_dict[TomoFluoConfig.DATASET_BASENAME_DK] == "test_"
    assert general_section_dict[TomoFluoConfig.TITLE_DK] == "my title"

    source_section_dict = config_dict[TomoFluoConfig.SOURCE_SECTION_DK]
    assert source_section_dict[TomoFluoConfig.INSTRUMENT_NAME_DK] == "BMXX"
    assert source_section_dict[TomoFluoConfig.SOURCE_NAME_DK] == "XFEL"
    assert (
        source_section_dict[TomoFluoConfig.SOURCE_TYPE_DK]
        == SourceType.FIXED_TUBE_X_RAY.value
    )



from __future__ import annotations

from copy import copy
from typing import Any
from silx.utils.enum import Enum as _Enum


__all__ = ["OptionLevel", "filter_options_level_items"]


class OptionLevel(_Enum):
    REQUIRED = "required"
    ADVANCED = "advanced"


def filter_options_level_items(
    dict_: dict[str, Any], level: OptionLevel, level_ref: dict[str, OptionLevel]
) -> dict[str, Any]:
    r"""
    Filer items in given dict from there configuration level.

    :param dict\_: that we want to filter. All possible keys must contained in `level_ref`. Else a KEyError exception will be raised.
    :param level: the OptionLevel we want to get. All keys which an higher option level will be removed from the dictionary
    :param level_ref: dict containing all possible 'dict\_' keys with associated option level
    """
    res = copy(dict_)
    keys = tuple(res.keys())
    for key in keys:
        if level == OptionLevel.REQUIRED and level_ref[key] == OptionLevel.ADVANCED:
            del res[key]
    return res



# coding: utf-8

import os
import pint
from tempfile import TemporaryDirectory

import pytest

from nxtomo.nxobject.nxdetector import FieldOfView
from nxtomo.nxobject.nxsource import ProbeType, SourceType
from nxtomomill import settings
from nxtomomill.io.config.edfconfig import TomoEDFConfig, generate_default_edf_config

_ureg = pint.get_application_registry()


def test_TomoEDFConfig_to_dict():
    """Test processing of the to_dict"""
    config = TomoEDFConfig()

    config_dict = config.to_dict()
    for key in (
        TomoEDFConfig.GENERAL_SECTION_DK,
        TomoEDFConfig.EDF_KEYS_SECTION_DK,
        TomoEDFConfig.FLAT_DARK_SECTION_DK,
        TomoEDFConfig.UNIT_SECTION_DK,
        TomoEDFConfig.SAMPLE_SECTION_DK,
        TomoEDFConfig.SOURCE_SECTION_DK,
        TomoEDFConfig.DETECTOR_SECTION_DK,
    ):
        assert key in config_dict

    TomoEDFConfig.from_dict(config_dict)

    with TemporaryDirectory() as folder:
        file_path = os.path.join(folder, "config.cfg")
        assert not os.path.exists(file_path)
        config.to_cfg_file(file_path)
        assert os.path.exists(file_path)
        config_loaded = TomoEDFConfig.from_cfg_file(file_path=file_path)

    assert config_loaded.to_dict() == config.to_dict()


def test_TomoEDFConfig_default_config():
    """insure default configuration generation works"""
    with TemporaryDirectory() as folder:
        file_path = os.path.join(folder, "config.cfg")
        assert not os.path.exists(file_path)
        config = generate_default_edf_config()
        assert isinstance(config, dict)
        TomoEDFConfig.from_dict(config).to_cfg_file(file_path=file_path)
        assert os.path.exists(file_path)


def test_TomoEDFConfig_setters():
    """test the different setters and getter of the EDFTomoConfig"""
    config = TomoEDFConfig()

    # try to test a new attribut (insure class is frozeen)
    with pytest.raises(AttributeError):
        config.new_attrs = "toto"

    # test general section setters
    with pytest.raises(TypeError):
        config.input_folder = 12.0
    config.input_folder = "my_folder"
    with pytest.raises(TypeError):
        config.output_file = 12.0
    config.output_file = "my_nx.nx"

    config.file_extension = ".nx"

    with pytest.raises(TypeError):
        config.dataset_basename = 12.0
    config.dataset_basename = None
    config.dataset_basename = "test_"

    with pytest.raises(TypeError):
        config.dataset_info_file = 12.3
    config.dataset_info_file = None
    config.dataset_info_file = "my_info_file.info"

    config.overwrite = True

    with pytest.raises(TypeError):
        config.title = 12.0
    config.title = None
    config.title = "my title"

    with pytest.raises(TypeError):
        config.ignore_file_patterns = 12.0
    with pytest.raises(TypeError):
        config.ignore_file_patterns = "toto"
    with pytest.raises(TypeError):
        config.ignore_file_patterns = (1.0,)
    config.ignore_file_patterns = ("toto",)
    config.ignore_file_patterns = None

    # test header keys + dark and flat setters
    attributes_value = {
        "motor_position_keys": settings.Tomo.EDF.MOTOR_POS,
        "motor_mne_keys": settings.Tomo.EDF.MOTOR_MNE,
        "rotation_angle_keys": settings.Tomo.EDF.ROT_ANGLE,
        "dark_names": settings.Tomo.EDF.DARK_NAMES,
        "flat_names": settings.Tomo.EDF.REFS_NAMES,
    }
    for attr, key in attributes_value.items():
        setattr(config, attr, key)
        with pytest.raises(TypeError):
            setattr(config, attr, "toto")
        with pytest.raises(TypeError):
            setattr(config, attr, 12.20)
        with pytest.raises(TypeError):
            setattr(config, attr, (12.20,))

    # test units setters
    attributes_value = {
        "pixel_size_unit": _ureg.micrometer,
        "distance_unit": _ureg.meter,
        "energy_unit": _ureg.keV,
        "x_trans_unit": _ureg.meter,
        "y_trans_unit": _ureg.meter,
        "z_trans_unit": _ureg.meter,
    }
    for attr, key in attributes_value.items():
        # test providing an instance of a unit
        setattr(config, attr, key)
        # test providing a sting if of a unit
        setattr(config, attr, str(key))
        with pytest.raises(TypeError):
            setattr(config, attr, None)
        with pytest.raises(TypeError):
            setattr(config, attr, 12.0)

    # test sample setters
    config.sample_name = None
    with pytest.raises(TypeError):
        config.sample_name = (12,)
    with pytest.raises(TypeError):
        config.sample_name = 12.0
    config.sample_name = "my sample"

    with pytest.raises(TypeError):
        config.force_angle_calculation = "toto"
    config.force_angle_calculation = True

    with pytest.raises(TypeError):
        config.force_angle_calculation_endpoint = "toto"
    config.force_angle_calculation_endpoint = True

    with pytest.raises(TypeError):
        config.angle_calculation_rev_neg_scan_range = "toto"
    config.angle_calculation_rev_neg_scan_range = False

    # test source setters
    config.instrument_name = None
    config.instrument_name = "BMXX"
    with pytest.raises(TypeError):
        config.instrument_name = 12.3

    config.source_name = None
    config.source_name = "ESRF"
    with pytest.raises(TypeError):
        config.source_name = 12.2
    config.source_name = "XFEL"

    config.source_type = None
    config.source_type = SourceType.FIXED_TUBE_X_RAY.value
    config.source_type = SourceType.FIXED_TUBE_X_RAY
    with pytest.raises(ValueError):
        config.source_type = "trsts"
    with pytest.raises(TypeError):
        config.source_type = 123

    config.source_probe = None
    config.source_probe = ProbeType.ELECTRON
    config.source_probe = "positron"
    with pytest.raises(TypeError):
        config.source_probe = 123

    # test detector setters
    config.field_of_view = None
    with pytest.raises(TypeError):
        config.field_of_view = 12
    with pytest.raises(ValueError):
        config.field_of_view = "toto"
    config.field_of_view = FieldOfView.FULL.value
    config.field_of_view = FieldOfView.HALF

    config_dict = config.to_dict()
    general_section_dict = config_dict[TomoEDFConfig.GENERAL_SECTION_DK]
    assert general_section_dict[TomoEDFConfig.INPUT_FOLDER_DK] == "my_folder"
    assert general_section_dict[TomoEDFConfig.OUTPUT_FILE_DK] == "my_nx.nx"
    assert general_section_dict[TomoEDFConfig.FILE_EXTENSION_DK] == ".nx"
    assert general_section_dict[TomoEDFConfig.OVERWRITE_DK] is True
    assert general_section_dict[TomoEDFConfig.DATASET_BASENAME_DK] == "test_"
    assert (
        general_section_dict[TomoEDFConfig.DATASET_FILE_INFO_DK] == "my_info_file.info"
    )
    assert general_section_dict[TomoEDFConfig.TITLE_DK] == "my title"
    assert general_section_dict[TomoEDFConfig.IGNORE_FILE_PATTERN_DK] == ""

    edf_headers_section_dict = config_dict[TomoEDFConfig.EDF_KEYS_SECTION_DK]
    assert (
        edf_headers_section_dict[TomoEDFConfig.MOTOR_POSITION_KEY_DK]
        == settings.Tomo.EDF.MOTOR_POS
    )
    assert (
        edf_headers_section_dict[TomoEDFConfig.MOTOR_MNE_KEY_DK]
        == settings.Tomo.EDF.MOTOR_MNE
    )
    assert (
        edf_headers_section_dict[TomoEDFConfig.X_TRANS_KEY_DK]
        == settings.Tomo.EDF.X_TRANS
    )
    assert (
        edf_headers_section_dict[TomoEDFConfig.Y_TRANS_KEY_DK]
        == settings.Tomo.EDF.Y_TRANS
    )
    assert (
        edf_headers_section_dict[TomoEDFConfig.Z_TRANS_KEY_DK]
        == settings.Tomo.EDF.Z_TRANS
    )
    assert (
        edf_headers_section_dict[TomoEDFConfig.ROT_ANGLE_KEY_DK]
        == settings.Tomo.EDF.ROT_ANGLE
    )

    dark_flat_section_dict = config_dict[TomoEDFConfig.FLAT_DARK_SECTION_DK]
    assert (
        dark_flat_section_dict[TomoEDFConfig.DARK_NAMES_DK]
        == settings.Tomo.EDF.DARK_NAMES
    )
    assert (
        dark_flat_section_dict[TomoEDFConfig.FLAT_NAMES_DK]
        == settings.Tomo.EDF.REFS_NAMES
    )

    units_section_dict = config_dict[TomoEDFConfig.UNIT_SECTION_DK]
    assert units_section_dict[TomoEDFConfig.PIXEL_SIZE_EXPECTED_UNIT] == str(
        _ureg.micrometer
    )
    assert units_section_dict[TomoEDFConfig.DISTANCE_EXPECTED_UNIT] == str(_ureg.meter)
    assert units_section_dict[TomoEDFConfig.ENERGY_EXPECTED_UNIT] == str(_ureg.keV)
    assert units_section_dict[TomoEDFConfig.X_TRANS_EXPECTED_UNIT] == str(_ureg.meter)
    assert units_section_dict[TomoEDFConfig.Y_TRANS_EXPECTED_UNIT] == str(_ureg.meter)
    assert units_section_dict[TomoEDFConfig.Z_TRANS_EXPECTED_UNIT] == str(_ureg.meter)

    sample_section_dict = config_dict[TomoEDFConfig.SAMPLE_SECTION_DK]
    assert sample_section_dict[TomoEDFConfig.SAMPLE_NAME_DK] == "my sample"
    assert sample_section_dict[TomoEDFConfig.FORCE_ANGLE_CALCULATION] is True
    assert sample_section_dict[TomoEDFConfig.FORCE_ANGLE_CALCULATION_ENDPOINT] is True
    assert (
        sample_section_dict[TomoEDFConfig.FORCE_ANGLE_CALCULATION_REVERT_NEG_SCAN_RANGE]
        is False
    )

    source_section_dict = config_dict[TomoEDFConfig.SOURCE_SECTION_DK]
    assert source_section_dict[TomoEDFConfig.INSTRUMENT_NAME_DK] == "BMXX"
    assert source_section_dict[TomoEDFConfig.SOURCE_NAME_DK] == "XFEL"
    assert (
        source_section_dict[TomoEDFConfig.SOURCE_TYPE_DK]
        == SourceType.FIXED_TUBE_X_RAY.value
    )

    detector_section_dict = config_dict[TomoEDFConfig.DETECTOR_SECTION_DK]
    assert (
        detector_section_dict[TomoEDFConfig.FIELD_OF_VIEW_DK] == FieldOfView.HALF.value
    )



# coding: utf-8

"""
contains the HDF5ConfigHandler
"""

from __future__ import annotations

import logging
import os

from nxtomomill import utils
from nxtomomill.io.utils import filter_str_def
from nxtomomill.converter.hdf5.utils import get_default_output_file

from .hdf5config import TomoHDF5Config

_logger = logging.getLogger(__name__)

__all__ = [
    "SETTABLE_PARAMETERS_UNITS",
    "SETTABLE_PARAMETERS_TYPE",
    "SETTABLE_PARAMETERS",
    "BaseHDF5ConfigHandler",
    "TomoHDF5ConfigHandler",
]

SETTABLE_PARAMETERS_UNITS = {
    "energy": "kev",
    "x_pixel_size": "m",
    "y_pixel_size": "m",
    "detector_sample_distance": "m",
}

SETTABLE_PARAMETERS_TYPE = {
    "energy": float,
    "x_pixel_size": float,
    "y_pixel_size": float,
    "detector_sample_distance": float,
}

SETTABLE_PARAMETERS = SETTABLE_PARAMETERS_UNITS.keys()


def _extract_param_value(key_values):
    """extract all the key / values elements from the str_list. Expected
    format is `param_1_name param_1_value param_2_name param_2_value ...`

    :param str_list: raw input string as `param_1_name param_1_value
                         param_2_name param_2_value ...`
    :return: dict of tuple (param_name, param_value)
    """
    if len(key_values) % 2 != 0:
        raise ValueError(
            "Expect a pair `param_name, param_value` for each " "parameters"
        )

    def pairwise(it):
        it = iter(it)
        while True:
            try:
                yield next(it), next(it)
            except StopIteration:
                # no more elements in the iterator
                return

    res = {}
    for name, value in pairwise(key_values):
        if name not in SETTABLE_PARAMETERS:
            raise ValueError(f"parameters {name} is not managed")
        if name in SETTABLE_PARAMETERS_TYPE:
            type_ = SETTABLE_PARAMETERS_TYPE[name]
            if type_ is not None:
                res[name] = type_(value)
                continue
        res[name] = value
    return res


class BaseHDF5ConfigHandler:
    """
    Class taking as input the CLI options and the user configuration file.
    Raises error if there is incoherence between the two (like if the CLI provides image_keys values AND if the configuration file provides the same input).

    This class will produce the instance of TomoHDF5Config to be used for the conversion
    """

    @staticmethod
    def get_tuple_of_keys_from_cmd(cmd_value):
        return utils.get_tuple_of_keys_from_cmd(cmd_value)

    @staticmethod
    def conv_str_to_bool(bstr):
        return bstr in ("True", True)

    @staticmethod
    def conv_log_level(bool_debug):
        if bool_debug is True:
            return "debug"
        else:
            return "warning"

    def __init__(self, argparse_options, raise_error=True):
        self._argparse_options = argparse_options
        self._config = None
        self.build_configuration(raise_error=raise_error)

    @property
    def configuration(self) -> TomoHDF5Config | None:
        return self._config

    @property
    def argparse_options(self):
        return self._argparse_options

    def _check_argparse_options(self, raise_error):
        raise NotImplementedError("BaseClass")

    def _create_HDF5_config(self):
        raise NotImplementedError("Base class")

    def build_configuration(self, raise_error) -> bool:
        """
        :param raise_error: raise error if encounter some errors. Else
                                 display a log message
        :return: True if the settings are valid
        """
        self._check_argparse_options(raise_error=raise_error)
        options = self.argparse_options
        config = self._create_HDF5_config()

        # check input and output file
        if config.input_file is None:
            config.input_file = options.input_file
        elif options.input_file is not None and config.input_file != options.input_file:
            raise ValueError(
                "Two different input files provided from "
                "command line and from the configuration file"
            )
        if config.input_file is None:
            err = "No input file provided"
            if raise_error:
                raise ValueError(err)
            else:
                _logger.error(err)

        if config.output_file is None:
            config.output_file = options.output_file
        elif (
            options.output_file is not None
            and config.output_file != options.output_file
        ):
            raise ValueError(
                "Two different output files provided from "
                "command line and from the configuration file"
            )
        if config.output_file is None:
            if config.file_extension is None:
                err = "If no output file provided you should provide the extension"
                raise ValueError(err)

            config.output_file = get_default_output_file(
                input_file=config.input_file,
                output_file_extension=config.file_extension.value,
            )
        elif os.path.isdir(config.output_file):
            # if the output file is only a directory: consider we want the same file basename with default extension
            # in this directory
            input_file, _ = os.path.splitext(config.input_file)
            config.output_file = os.path.join(
                os.path.abspath(config.output_file),
                os.path.basename(input_file) + config.file_extension.value,
            )

        # set parameter from the arg parse options
        # key is the name of the argparse option.
        # value is a tuple: (name of the setter in the HDF5Config,
        # function to format the input)
        self._config = self._map_option_to_config_param(config, options)

    def _map_option_to_config_param(self, config, options):
        raise NotImplementedError("Base class")

    def __str__(self):
        raise NotImplementedError("")


class TomoHDF5ConfigHandler(BaseHDF5ConfigHandler):
    """
    Create a TomoHDF5Config from argparse option provided from the CLI.
    In the case a configuration is provided it also check that a parameter is not defined twice (from the CLI and from the configuration file)
    """

    def _create_HDF5_config(self):
        if self.argparse_options.config:
            return TomoHDF5Config.from_cfg_file(self.argparse_options.config)
        else:
            return TomoHDF5Config()

    def _map_option_to_config_param(self, config, options):
        mapping = {
            "valid_camera_names": (
                "valid_camera_names",
                self.get_tuple_of_keys_from_cmd,
            ),
            "overwrite": ("overwrite", self.conv_str_to_bool),
            "file_extension": ("file_extension", filter_str_def),
            "single_file": ("single_file", self.conv_str_to_bool),
            "no_master_file": ("no_master_file", self.conv_str_to_bool),
            "debug": ("log_level", self.conv_log_level),
            "entries": ("entries", self.get_tuple_of_keys_from_cmd),
            "ignore_sub_entries": (
                "sub_entries_to_ignore",
                self.get_tuple_of_keys_from_cmd,
            ),
            "duplicate_data": ("default_copy_behavior", self.conv_str_to_bool),
            "raises_error": ("raises_error", self.conv_str_to_bool),
            "field_of_view": ("field_of_view", filter_str_def),
            "request_input": ("request_input", self.conv_str_to_bool),
            "sample_x_keys": ("sample_x_keys", self.get_tuple_of_keys_from_cmd),
            "sample_y_keys": ("sample_y_keys", self.get_tuple_of_keys_from_cmd),
            "translation_z_keys": (
                "translation_z_keys",
                self.get_tuple_of_keys_from_cmd,
            ),
            "rot_angle_keys": ("rotation_angle_keys", self.get_tuple_of_keys_from_cmd),
            "sample_detector_distance_paths": (
                "sample_detector_distance_paths",
                self.get_tuple_of_keys_from_cmd,
            ),
            "acq_expo_time_keys": (
                "exposition_time_keys",
                self.get_tuple_of_keys_from_cmd,
            ),
            "x_pixel_size_key": ("x_pixel_size_paths", self.get_tuple_of_keys_from_cmd),
            "y_pixel_size_key": ("y_pixel_size_paths", self.get_tuple_of_keys_from_cmd),
            "init_titles": ("init_titles", self.get_tuple_of_keys_from_cmd),
            "init_zserie_titles": (
                "zserie_init_titles",
                self.get_tuple_of_keys_from_cmd,
            ),
            "init_multi_tomo_titles": (
                "multitomo_init_titles",
                self.get_tuple_of_keys_from_cmd,
            ),
            "init_back_and_forth_titles": (
                "back_and_forth_init_titles",
                self.get_tuple_of_keys_from_cmd,
            ),
            "dark_titles": ("dark_titles", self.get_tuple_of_keys_from_cmd),
            "flat_titles": ("flat_titles", self.get_tuple_of_keys_from_cmd),
            "proj_titles": ("projections_titles", self.get_tuple_of_keys_from_cmd),
            "align_titles": ("alignment_titles", self.get_tuple_of_keys_from_cmd),
            "set_params": ("param_already_defined", _extract_param_value),
        }
        for argparse_name, (config_name, format_fct) in mapping.items():
            argparse_value = getattr(options, argparse_name)
            if argparse_value is not None:
                value = format_fct(argparse_value)  # pylint: disable=E1102
                setattr(config, config_name, value)
        return config

    def _check_argparse_options(self, raise_error):
        if self.argparse_options is None:
            err = "No argparse options provided"
            if raise_error:
                raise ValueError(err)
            else:
                _logger.error(err)
            return False

        options = self.argparse_options
        if options.config is not None:
            # check no other option are provided
            duplicated_inputs = []
            for opt in (
                "set_params",
                "align_titles",
                "proj_titles",
                "flat_titles",
                "dark_titles",
                "init_zserie_titles",
                "init_titles",
                "init_multi_tomo_titles",
                "x_pixel_size_key",
                "y_pixel_size_key",
                "acq_expo_time_keys",
                "rot_angle_keys",
                "valid_camera_names",
                "translation_z_keys",
                "sample_y_keys",
                "sample_x_keys",
                "request_input",
                "raises_error",
                "ignore_sub_entries",
                "entries",
                "debug",
                "overwrite",
                "single_file",
                "no_master_file",
                "file_extension",
                "field_of_view",
                "duplicate_data",
            ):
                if getattr(options, opt):
                    duplicated_inputs.append(opt)
            if len(duplicated_inputs) > 0:
                err = f"You provided a configuration file and inputs for {duplicated_inputs}"
                if raise_error:
                    raise ValueError(err)
                else:
                    _logger.error(err)
                return False



# coding: utf-8

import os
import tempfile

from nxtomomill.app.h52nx import main
from nxtomomill.tests.utils.bliss import MockBlissAcquisition


def test_h52nx_application():
    """test nxtomomill h52nx input_file output_file --single-file"""
    with tempfile.TemporaryDirectory() as folder:
        nx_file_path = os.path.join(folder, "acquisition.nx")

        bliss_mock = MockBlissAcquisition(
            n_sample=1,
            n_sequence=1,
            n_scan_per_sequence=10,
            n_darks=5,
            n_flats=5,
            with_nx_detector_attr=True,
            output_dir=folder,
            detector_name="pcolinux",
        )

        h5_file_path = bliss_mock.samples[0].sample_file
        assert not os.path.exists(nx_file_path), "outputfile exists already"
        main(["h52nx", h5_file_path, nx_file_path, "--single-file"])
        assert os.path.exists(nx_file_path), "outputfile doesn't exists"



# coding: utf-8

from __future__ import annotations

import configparser
import logging

from nxtomo.nxobject.nxsource import ProbeType, SourceType

from nxtomomill.io.config.optionlevel import OptionLevel, filter_options_level_items
from nxtomomill.io.config.configbase import ConfigBase, ConfigSourceSection
from nxtomomill.io.utils import (
    PathType,
    convert_str_to_bool,
    convert_str_to_tuple,
    filter_str_def,
)
from nxtomomill.settings import Tomo
from nxtomomill.utils import FileExtension

_logger = logging.getLogger(__name__)


__all__ = ["TomoFluoConfig", "generate_default_fluo_config"]


class TomoFluoConfig(ConfigBase, ConfigSourceSection):
    """
    Configuration class to provide to the fluo->nx converter .
    """

    # General section keys

    GENERAL_SECTION_DK = "GENERAL_SECTION"

    INPUT_FOLDER_DK = "input_folder"

    OUTPUT_FILE_DK = "output_file"

    DETECTOR_NAMES_DK = "detector_names"

    DIMENSION_DK = "dimension"

    FILE_EXTENSION_DK = "file_extension"

    OVERWRITE_DK = "overwrite"

    DATASET_BASENAME_DK = "dataset_basename"

    DATASET_FILE_INFO_DK = "dataset_info_file"

    LOG_LEVEL_DK = "log_level"

    TITLE_DK = "title"

    IGNORE_FILE_PATTERN_DK = "patterns_to_ignores"

    DUPLICATE_DATA_DK = "duplicate_data"

    EXTERNAL_LINK_RELATIVE_DK = "external_link_path"

    COMMENTS_GENERAL_SECTION = {
        GENERAL_SECTION_DK: "general information. \n",
        INPUT_FOLDER_DK: "Path to the folder containing the raw data folder and the fluofit/ subfolder. if not provided from the configuration file must be provided from the command line",
        OUTPUT_FILE_DK: "File produced by the converter. '.nx' extension recommended. If not provided from the configuration file must be provided from the command line",
        DIMENSION_DK: "Dimension of the experiment. 2 for 2D XRFCT, 3 for 3D XRFCT. Default is 3.",
        DETECTOR_NAMES_DK: "Define a list of (real or virtual) detector names used for the exp (space separated values - no comma). E.g. 'falcon xmap'. If not specified, all detectors are processed.",
        OVERWRITE_DK: "overwrite output files if exists without asking",
        FILE_EXTENSION_DK: "file extension. Ignored if the output file is provided and contains an extension",
        DATASET_BASENAME_DK: f"In 2D, the exact full name of the folder. In 3D, the folder name prefix (the program will search for folders named <prefix>_XXX where XXX is a nmber.) If not provided will take the name of {INPUT_FOLDER_DK}",
        DATASET_FILE_INFO_DK: f"path to .info file containing dataset information (Energy, ScanRange, TOMO_N...). If not will deduce it from {DATASET_BASENAME_DK}",
        LOG_LEVEL_DK: 'Log level. Valid levels are "debug", "info", "warning" and "error"',
        TITLE_DK: "NXtomo title",
        IGNORE_FILE_PATTERN_DK: "some file pattern leading to ignoring the file. Like reconstructed slice files.",
        DUPLICATE_DATA_DK: "If False then will create embed all the data into a single file avoiding external link to other file. If True then the decetor data will point to original tif files. In this case you must be carreful to keep relative paths valid. Warning: to read external dataset you nust be at the hdf5 file working directory. See external link resolution details: https://support.hdfgroup.org/documentation/hdf5/latest/group___h5_l.html#title5",
        EXTERNAL_LINK_RELATIVE_DK: "If 'duplicate_data' is set to False then you can specify if you want the link to original files to be 'relative' or 'absolute'",
    }

    LEVEL_GENERAL_SECTION = {
        INPUT_FOLDER_DK: OptionLevel.REQUIRED,
        OUTPUT_FILE_DK: OptionLevel.REQUIRED,
        DIMENSION_DK: OptionLevel.REQUIRED,
        DETECTOR_NAMES_DK: OptionLevel.REQUIRED,
        OVERWRITE_DK: OptionLevel.REQUIRED,
        FILE_EXTENSION_DK: OptionLevel.ADVANCED,
        DATASET_BASENAME_DK: OptionLevel.REQUIRED,
        DATASET_FILE_INFO_DK: OptionLevel.REQUIRED,
        LOG_LEVEL_DK: OptionLevel.REQUIRED,
        TITLE_DK: OptionLevel.ADVANCED,
        IGNORE_FILE_PATTERN_DK: OptionLevel.REQUIRED,
        DUPLICATE_DATA_DK: OptionLevel.REQUIRED,
        EXTERNAL_LINK_RELATIVE_DK: OptionLevel.ADVANCED,
    }

    # create comments

    COMMENTS = COMMENTS_GENERAL_SECTION
    COMMENTS.update(ConfigSourceSection.COMMENTS_SOURCE_SECTION_DK)

    SECTIONS_LEVEL = {
        GENERAL_SECTION_DK: OptionLevel.REQUIRED,
        ConfigSourceSection.SOURCE_SECTION_DK: OptionLevel.ADVANCED,
    }

    def __init__(self):
        super().__init__()
        self._set_freeze(False)
        # general information
        self._input_folder = None
        self._output_file = None
        self._dimension: int = 3
        self._detectors: tuple[str] = tuple()
        self._file_extension = FileExtension.NX
        self._overwrite = False
        self._dataset_basename = None
        self._log_level = logging.WARNING
        self._title = None
        self._ignore_file_patterns = Tomo.EDF.TO_IGNORE
        self._dataset_info_file = None
        self._duplicate_data = True
        self._external_path_type = PathType.RELATIVE

        # source
        self._instrument_name = None
        self._source_name = "ESRF"
        self._source_type = SourceType.SYNCHROTRON_X_RAY_SOURCE
        self._source_probe = ProbeType.X_RAY

        self._set_freeze(True)

    @property
    def input_folder(self) -> str | None:
        return self._input_folder

    @input_folder.setter
    def input_folder(self, folder: str | None) -> None:
        if not isinstance(folder, (type(None), str)):
            raise TypeError(
                f"folder is expected to be None or an instance of str. Not {type(folder)}"
            )
        self._input_folder = folder

    @property
    def dimension(self) -> int:
        return self._dimension

    @dimension.setter
    def dimension(self, dimension: int):
        self._dimension = int(dimension)

    @property
    def detectors(self) -> list:

        return self._detectors

    @detectors.setter
    def detectors(self, dets: tuple | str):
        self._detectors = convert_str_to_tuple(dets)

    @property
    def dataset_basename(self) -> str | None:
        return self._dataset_basename

    @dataset_basename.setter
    def dataset_basename(self, dataset_basename: str | None) -> None:
        if not isinstance(dataset_basename, (type(None), str)):
            raise TypeError(
                f"Dataset_basename is expected to be None or an instance of str. Not {type(dataset_basename)}"
            )
        self._dataset_basename = dataset_basename

    @property
    def dataset_info_file(self) -> str | None:
        return self._dataset_info_file

    @dataset_info_file.setter
    def dataset_info_file(self, file_path: str | None) -> None:
        if not isinstance(file_path, (type(None), str)):
            raise TypeError(
                f"info.file path is expected to be None or an instance of str. Not {type(file_path)}"
            )
        self._dataset_info_file = file_path

    @property
    def duplicate_data(self) -> bool:
        return self._duplicate_data

    @duplicate_data.setter
    def duplicate_data(self, duplicate: bool):
        if not isinstance(duplicate, bool):
            raise TypeError(
                f"Duplicate is expected to be a bool and not {type(duplicate)}"
            )
        self._duplicate_data = duplicate

    @property
    def external_path_type(self) -> PathType:
        return self._external_path_type

    @external_path_type.setter
    def external_path_type(self, path_type: str | PathType):
        self._external_path_type = PathType(path_type)

    @property
    def title(self) -> str | None:
        return self._title

    @title.setter
    def title(self, title: str | None) -> None:
        if not isinstance(title, (type(None), str)):
            raise TypeError(
                f"title is expected to be None or an instance of str. Not {type(title)}"
            )
        self._title = title

    @property
    def ignore_file_patterns(self) -> tuple:
        return self._ignore_file_patterns

    @ignore_file_patterns.setter
    def ignore_file_patterns(self, patterns: tuple | list | None):
        if not isinstance(patterns, (type(None), tuple, list)):
            raise TypeError("patterns is expected to be a tuple or a list")
        if patterns is None:
            self._ignore_file_patterns = tuple()
        else:
            for elmt in patterns:
                if not isinstance(elmt, str):
                    raise TypeError("patterns elmts are expected to be str")
            self._ignore_file_patterns = tuple(patterns)

    @property
    def instrument_name(self) -> str | None:
        return self._instrument_name

    @instrument_name.setter
    def instrument_name(self, name: str | None):
        if not isinstance(name, (type(None), str)):
            raise TypeError("name is expected to be None or an instance of str")
        self._instrument_name = name

    @property
    def source_name(self) -> str | None:
        return self._source_name

    @source_name.setter
    def source_name(self, name: str | None) -> None:
        if not isinstance(name, (type(None), str)):
            raise TypeError("name is expected to be None or an instance of str")
        self._source_name = name

    @property
    def source_type(self) -> SourceType | None:
        return self._source_type

    @source_type.setter
    def source_type(self, source_type: SourceType | str | None):
        if not isinstance(source_type, (type(None), str, SourceType)):
            raise TypeError(
                "source_type is expected to be None or an instance of SourceType or str"
            )
        if source_type is None:
            self._source_type = None
        else:
            self._source_type = SourceType(source_type)

    @property
    def source_probe(self) -> ProbeType | None:
        return self._source_probe

    @source_probe.setter
    def source_probe(self, source_probe: ProbeType | str | None):
        if not isinstance(source_probe, (type(None), str, ProbeType)):
            raise TypeError(
                "source_probe is expected to be None or an instance of ProbeType or str"
            )
        if source_probe is None:
            self._source_probe = None
        else:
            self._source_probe = ProbeType(source_probe)

    def to_dict(self, level="advanced") -> dict:
        """convert the configuration to a dictionary"""
        level = OptionLevel(level)
        sections_callback = {
            self.GENERAL_SECTION_DK: self._general_section_to_dict,
            self.SOURCE_SECTION_DK: self._source_section_to_dict,
        }
        res = {}
        for section, callback in sections_callback.items():
            if (
                level == OptionLevel.ADVANCED
                or TomoFluoConfig.SECTIONS_LEVEL[section] == OptionLevel.REQUIRED
            ):
                res[section] = callback(level=level)
        return res

    def _general_section_to_dict(self, level) -> dict:
        res = {
            self.INPUT_FOLDER_DK: (
                self.input_folder if self.input_folder is not None else ""
            ),
            self.OUTPUT_FILE_DK: (
                self.output_file if self.output_file is not None else ""
            ),
            self.DIMENSION_DK: (self.dimension if self.dimension is not None else ""),
            self.DETECTOR_NAMES_DK: (
                self.detectors if self.detectors is not None else ""
            ),
            self.OVERWRITE_DK: self.overwrite,
            self.FILE_EXTENSION_DK: self.file_extension.value,
            self.DATASET_BASENAME_DK: (
                self.dataset_basename if self.dataset_basename is not None else ""
            ),
            self.DATASET_FILE_INFO_DK: (
                self.dataset_info_file if self.dataset_info_file is not None else ""
            ),
            self.LOG_LEVEL_DK: logging.getLevelName(self.log_level).lower(),
            self.TITLE_DK: self.title if self.title is not None else "",
            self.IGNORE_FILE_PATTERN_DK: (
                self.ignore_file_patterns
                if self.ignore_file_patterns != tuple()
                else ""
            ),
            self.DUPLICATE_DATA_DK: self.duplicate_data,
            self.EXTERNAL_LINK_RELATIVE_DK: self.external_path_type.value,
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=TomoFluoConfig.LEVEL_GENERAL_SECTION
        )

    @staticmethod
    def from_dict(dict_: dict):
        r"""
        Create a HDF5Config object and set it from values contained in the
        dictionary
        :param dict\_: settings dictionary
        :return: HDF5Config
        """
        config = TomoFluoConfig()
        config.load_from_dict(dict_)
        return config

    def load_from_dict(self, dict_: dict) -> None:
        """Load the configuration from a dictionary"""
        sections_loaders = {
            TomoFluoConfig.GENERAL_SECTION_DK: self.load_general_section,
            TomoFluoConfig.SOURCE_SECTION_DK: self.load_source_section,
        }
        for section_key, loaded_func in sections_loaders.items():
            if section_key in dict_:
                loaded_func(dict_[section_key])
            else:
                _logger.info(
                    f"No {section_key} section found. Will take default values"
                )

    def load_general_section(self, dict_: dict) -> None:
        self.input_folder = dict_.get(TomoFluoConfig.INPUT_FOLDER_DK, None)
        self.output_file = dict_.get(TomoFluoConfig.OUTPUT_FILE_DK, None)
        self.detectors = dict_.get(TomoFluoConfig.DETECTOR_NAMES_DK, None)
        self.dimension = dict_.get(TomoFluoConfig.DIMENSION_DK, 3)
        if self.detectors is not None:
            if self.detectors == "":
                self.detectors = None

        overwrite = dict_.get(TomoFluoConfig.OVERWRITE_DK, None)
        if overwrite is not None:
            self.overwrite = convert_str_to_bool(overwrite)

        file_extension = dict_.get(TomoFluoConfig.FILE_EXTENSION_DK, None)
        if file_extension not in (None, ""):
            self.file_extension = filter_str_def(file_extension)

        dataset_basename = dict_.get(TomoFluoConfig.DATASET_BASENAME_DK, None)
        if dataset_basename is not None:
            if dataset_basename == "":
                dataset_basename = None
            self.dataset_basename = dataset_basename

        dataset_info_file = dict_.get(TomoFluoConfig.DATASET_FILE_INFO_DK, None)
        if dataset_info_file is not None:
            if dataset_info_file == "":
                dataset_info_file = None
            self.dataset_info_file = dataset_info_file

        log_level = dict_.get(TomoFluoConfig.LOG_LEVEL_DK, None)
        if log_level is not None:
            self.log_level = log_level
        self.title = dict_.get(TomoFluoConfig.TITLE_DK)
        ignore_file_patterns = dict_.get(TomoFluoConfig.IGNORE_FILE_PATTERN_DK, None)
        if ignore_file_patterns is not None:
            if ignore_file_patterns == "":
                ignore_file_patterns = tuple()
            else:
                ignore_file_patterns = convert_str_to_tuple(ignore_file_patterns)
            self.ignore_file_patterns = ignore_file_patterns
        duplicate_data = dict_.get(TomoFluoConfig.DUPLICATE_DATA_DK, None)
        if duplicate_data is not None:
            self.duplicate_data = convert_str_to_bool(duplicate_data)
        external_path_type = dict_.get(TomoFluoConfig.EXTERNAL_LINK_RELATIVE_DK, None)
        if external_path_type is not None:
            self.external_path_type = external_path_type

    def to_cfg_file(self, file_path: str):
        # TODO: add some generic information like:provided order of the tuple
        # will be the effective one. You can provide a key from it names if
        # it is contained in the positioners group
        # maybe split in sub section ?
        self.dict_to_cfg(file_path=file_path, dict_=self.to_dict())

    @staticmethod
    def dict_to_cfg(file_path, dict_):
        """ """
        return ConfigBase._dict_to_cfg(
            file_path=file_path,
            dict_=dict_,
            comments_fct=TomoFluoConfig.get_comments,
            logger=_logger,
        )

    @staticmethod
    def from_cfg_file(file_path: str, encoding=None):
        assert file_path is not None, "file_path should not be None"
        config_parser = configparser.ConfigParser(allow_no_value=True)
        config_parser.read(file_path, encoding=encoding)
        return TomoFluoConfig.from_dict(config_parser)

    @staticmethod
    def get_comments(key):
        return TomoFluoConfig.COMMENTS[key]


def generate_default_fluo_config(level: str = "required") -> dict:
    """generate a default configuration to convert fluo-tomo data (after PyMCA fit) to NXtomo"""
    return TomoFluoConfig().to_dict(level=level)



# coding: utf-8

from __future__ import annotations
import configparser
import logging
from typing import Iterable


from nxtomomill.io.config.optionlevel import OptionLevel, filter_options_level_items
from nxtomo.nxobject.nxdetector import FieldOfView
from nxtomomill.utils import FileExtension
from nxtomo.nxobject.nxsource import ProbeType, SourceType

__all__ = ["ConfigBase", "ConfigSourceSection"]


class ConfigSourceSection:
    # SOURCE SECTION

    SOURCE_SECTION_DK = "SOURCE_SECTION"

    INSTRUMENT_NAME_DK = "instrument_name"

    SOURCE_NAME_DK = "source_name"

    SOURCE_TYPE_DK = "source_type"

    SOURCE_PROBE_DK = "source_probe"

    COMMENTS_SOURCE_SECTION_DK = {
        SOURCE_SECTION_DK: "section dedicated to source definition.\n",
        INSTRUMENT_NAME_DK: "name of the instrument",
        SOURCE_NAME_DK: "name of the source",
        SOURCE_TYPE_DK: f"type of the source. Must be one of {[item.value for item in SourceType]}",
        SOURCE_PROBE_DK: f"source probe. Must be one of {[item.value for item in ProbeType]}",
    }

    LEVEL_SOURCE_SECTION = {
        INSTRUMENT_NAME_DK: OptionLevel.ADVANCED,
        SOURCE_NAME_DK: OptionLevel.ADVANCED,
        SOURCE_TYPE_DK: OptionLevel.ADVANCED,
        SOURCE_PROBE_DK: OptionLevel.ADVANCED,
    }

    def load_source_section(self, dict_: dict) -> None:
        if self.INSTRUMENT_NAME_DK in dict_:
            self.instrument_name = dict_.get(self.INSTRUMENT_NAME_DK)
        if self.SOURCE_NAME_DK in dict_:
            self.source_name = dict_.get(self.SOURCE_NAME_DK)
        if self.SOURCE_TYPE_DK in dict_:
            self.source_type = dict_[self.SOURCE_TYPE_DK]
        if self.SOURCE_PROBE_DK in dict_:
            self.source_probe = dict_[self.SOURCE_PROBE_DK]

    def _source_section_to_dict(self, level) -> dict:
        res = {
            self.INSTRUMENT_NAME_DK: self.instrument_name or "",
            self.SOURCE_NAME_DK: self.source_name or "",
            self.SOURCE_TYPE_DK: (
                self.source_type.value if self.source_type is not None else ""
            ),
            self.SOURCE_PROBE_DK: (
                self.source_probe.value if self.source_probe is not None else ""
            ),
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=self.LEVEL_SOURCE_SECTION
        )


class ConfigBase:
    __isfrozen = False
    # to ease API and avoid setting wrong attributes we 'freeze' the attributes
    # see https://stackoverflow.com/questions/3603502/prevent-creating-new-attributes-outside-init

    def __init__(self) -> None:
        self._output_file = None
        self._overwrite = False
        self._file_extension = FileExtension.NX
        self._log_level = logging.WARNING
        self._field_of_view = None
        self._machine_current_keys = None

    def __setattr__(self, __name, __value):
        if self.__isfrozen and not hasattr(self, __name):
            raise AttributeError("can't set attribute", __name)
        else:
            super().__setattr__(__name, __value)

    @property
    def output_file(self) -> str | None:
        return self._output_file

    @output_file.setter
    def output_file(self, output_file: str | None):
        if not isinstance(output_file, (str, type(None))):
            raise TypeError("'input_file' should be None or an instance of Iterable")
        elif output_file == "":
            self._output_file = None
        else:
            self._output_file = output_file

    @property
    def overwrite(self) -> bool:
        return self._overwrite

    @overwrite.setter
    def overwrite(self, overwrite: bool) -> None:
        if not isinstance(overwrite, bool):
            raise TypeError("'overwrite' should be a boolean")
        else:
            self._overwrite = overwrite

    @property
    def file_extension(self) -> FileExtension:
        return self._file_extension

    @file_extension.setter
    def file_extension(self, file_extension: str):
        self._file_extension = FileExtension(file_extension)

    @property
    def log_level(self):
        return self._log_level

    @log_level.setter
    def log_level(self, level: str):
        self._log_level = getattr(logging, level.upper())

    def _set_freeze(self, freeze=True):
        self.__isfrozen = freeze

    @property
    def field_of_view(self) -> FieldOfView | None:
        return self._field_of_view

    @field_of_view.setter
    def field_of_view(self, fov: FieldOfView | str | None):
        if fov is None:
            self._field_of_view = fov
        elif isinstance(fov, str):
            self._field_of_view = FieldOfView(fov.title())
        elif isinstance(fov, FieldOfView):
            self._field_of_view = fov
        else:
            raise TypeError(
                f"fov is expected to be None, a string or FieldOfView. Not {type(fov)}"
            )

    @property
    def rotation_angle_keys(self) -> Iterable:
        return self._rot_angle_keys

    @rotation_angle_keys.setter
    def rotation_angle_keys(self, keys: Iterable):
        if not isinstance(keys, Iterable) or isinstance(keys, str):
            raise TypeError("'keys' should be an Iterable")
        else:
            for elmt in keys:
                if not isinstance(elmt, str):
                    raise TypeError("keys elmts are expected to be str")
            self._rot_angle_keys = keys

    @property
    def translation_z_keys(self) -> Iterable:
        return self._translation_z_keys

    @translation_z_keys.setter
    def translation_z_keys(self, keys) -> None:
        if not isinstance(keys, Iterable) or isinstance(keys, str):
            raise TypeError("'keys' should be an Iterable")
        else:
            for elmt in keys:
                if not isinstance(elmt, str):
                    raise TypeError("keys elmts are expected to be str")
            self._translation_z_keys = keys

    @property
    def machine_current_keys(self) -> Iterable:
        return self._machine_current_keys

    @machine_current_keys.setter
    def machine_current_keys(self, keys: Iterable) -> None:
        self._machine_current_keys = keys

    def to_dict(self) -> dict:
        """convert the configuration to a dictionary"""
        raise NotImplementedError("Base class")

    def load_from_dict(self, dict_: dict) -> None:
        """Load the configuration from a dictionary"""
        raise NotImplementedError("Base class")

    @staticmethod
    def from_dict(dict_: dict):
        raise NotImplementedError("Base class")

    def to_cfg_file(self, file_path: str):
        # TODO: add some generic information like:provided order of the tuple
        # will be the effective one. You can provide a key from it names if
        # it is contained in the positioners group
        # maybe split in sub section ?
        self.dict_to_cfg(file_path=file_path, dict_=self.to_dict())

    @staticmethod
    def dict_to_cfg(file_path, dict_):
        """ """
        raise NotImplementedError("Base class")

    @staticmethod
    def _dict_to_cfg(file_path, dict_, comments_fct, logger):
        """ """
        if not file_path.lower().endswith((".cfg", ".config", ".conf")):
            logger.warning("add a valid extension to the output file")
            file_path += ".cfg"
        config = configparser.ConfigParser(allow_no_value=True)
        config.optionxform = str
        for section_name, values in dict_.items():
            config.add_section(section_name)
            config.set(section_name, "# " + comments_fct(section_name), None)
            for key, value in values.items():
                # adopt nabu design: comments are set prior to the key
                config.set(section_name, "# " + comments_fct(key), None)
                config.set(section_name, key, str(value))

        with open(file_path, "w") as config_file:
            config.write(config_file)

    @staticmethod
    def from_cfg_file(file_path: str, encoding=None):
        raise NotImplementedError("Base class")

    @staticmethod
    def get_comments(key):
        raise NotImplementedError("Base class")

    def __str__(self):
        return str(self.to_dict())



# coding: utf-8

import os
import shutil
import tempfile
import unittest

from nxtomomill import settings
from nxtomomill.io.config import TomoHDF5Config


class TestH5Config(unittest.TestCase):
    """
    Test the HDF5Config class
    """

    def setUp(self) -> None:
        self.folder = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)

    def test_generate_default_config(self):
        """
        Insure we can generate a default configuration
        """
        config = TomoHDF5Config()
        config.input_file = "toto.h5"
        config.output_file = "toto.nx"
        output = config.to_dict()

        self.assertEqual(type(output), dict)
        # check titles values
        titles_dict = output[TomoHDF5Config.ENTRIES_AND_TITLES_SECTION_DK]
        self.assertEqual(
            titles_dict[TomoHDF5Config.INIT_TITLES_DK], settings.Tomo.H5.INIT_TITLES
        )
        self.assertEqual(
            titles_dict[TomoHDF5Config.ZSERIE_INIT_TITLES_DK],
            settings.Tomo.H5.ZSERIE_INIT_TITLES,
        )
        self.assertEqual(
            titles_dict[TomoHDF5Config.PROJ_TITLES_DK], settings.Tomo.H5.PROJ_TITLES
        )
        self.assertEqual(
            titles_dict[TomoHDF5Config.FLAT_TITLES_DK], settings.Tomo.H5.FLAT_TITLES
        )
        self.assertEqual(
            titles_dict[TomoHDF5Config.DARK_TITLES_DK], settings.Tomo.H5.DARK_TITLES
        )
        self.assertEqual(
            titles_dict[TomoHDF5Config.ALIGNMENT_TITLES_DK],
            settings.Tomo.H5.ALIGNMENT_TITLES,
        )
        # check sample pixel size
        keys_dict = output[TomoHDF5Config.KEYS_SECTION_DK]
        self.assertEqual(
            keys_dict[TomoHDF5Config.SAMPLE_X_PIXEL_SIZE_KEYS_DK],
            settings.Tomo.H5.SAMPLE_X_PIXEL_SIZE_KEYS,
        )
        self.assertEqual(
            keys_dict[TomoHDF5Config.SAMPLE_Y_PIXEL_SIZE_KEYS_DK],
            settings.Tomo.H5.SAMPLE_Y_PIXEL_SIZE_KEYS,
        )
        self.assertEqual(
            keys_dict[TomoHDF5Config.DETECTOR_X_PIXEL_SIZE_KEYS_DK],
            settings.Tomo.H5.DETECTOR_X_PIXEL_SIZE_KEYS,
        )
        self.assertEqual(
            keys_dict[TomoHDF5Config.DETECTOR_Y_PIXEL_SIZE_KEYS_DK],
            settings.Tomo.H5.DETECTOR_Y_PIXEL_SIZE_KEYS,
        )
        # check translation
        self.assertEqual(
            keys_dict[TomoHDF5Config.SAMPLE_X_KEYS_DK], settings.Tomo.H5.SAMPLE_X_KEYS
        )
        self.assertEqual(
            keys_dict[TomoHDF5Config.SAMPLE_Y_KEYS_DK], settings.Tomo.H5.SAMPLE_Y_KEYS
        )
        self.assertEqual(
            keys_dict[TomoHDF5Config.TRANSLATION_Z_KEYS_DK],
            settings.Tomo.H5.TRANSLATION_Z_KEYS,
        )
        # others
        if settings.Tomo.H5.VALID_CAMERA_NAMES is None:
            self.assertEqual(keys_dict[TomoHDF5Config.VALID_CAMERA_DK], "")
        else:
            self.assertEqual(
                keys_dict[TomoHDF5Config.VALID_CAMERA_DK],
                settings.Tomo.H5.VALID_CAMERA_NAMES,
            )
        self.assertEqual(
            keys_dict[TomoHDF5Config.ROT_ANGLE_DK], settings.Tomo.H5.ROT_ANGLE_KEYS
        )
        self.assertEqual(
            keys_dict[TomoHDF5Config.DIODE_KEYS_DK], settings.Tomo.H5.DIODE_KEYS
        )
        self.assertEqual(
            keys_dict[TomoHDF5Config.TRANSLATION_Y_KEYS_DK],
            settings.Tomo.H5.TRANSLATION_Y_KEYS,
        )
        self.assertEqual(
            keys_dict[TomoHDF5Config.ACQUISITION_EXPO_TIME_KEYS_DK],
            settings.Tomo.H5.ACQ_EXPO_TIME_KEYS,
        )

        # check input and output file
        general_information = output[TomoHDF5Config.GENERAL_SECTION_DK]
        self.assertEqual(general_information[TomoHDF5Config.INPUT_FILE_DK], "toto.h5")
        self.assertEqual(general_information[TomoHDF5Config.OUTPUT_FILE_DK], "toto.nx")

    def test_to_dict(self):
        """test the `to_dict` function"""
        config = TomoHDF5Config()
        output_dict = config.to_dict()
        self.assertEqual(type(output_dict), dict)
        # check sections
        for section in (
            TomoHDF5Config.GENERAL_SECTION_DK,
            TomoHDF5Config.KEYS_SECTION_DK,
            TomoHDF5Config.EXTRA_PARAMS_SECTION_DK,
            TomoHDF5Config.FRAME_TYPE_SECTION_DK,
            TomoHDF5Config.ENTRIES_AND_TITLES_SECTION_DK,
        ):
            with self.subTest(section=section):
                self.assertTrue(section in output_dict)
        # check titles keys
        for key in (
            TomoHDF5Config.ALIGNMENT_TITLES_DK,
            TomoHDF5Config.PROJ_TITLES_DK,
            TomoHDF5Config.ZSERIE_INIT_TITLES_DK,
            TomoHDF5Config.INIT_TITLES_DK,
            TomoHDF5Config.FLAT_TITLES_DK,
            TomoHDF5Config.DARK_TITLES_DK,
        ):
            with self.subTest(key=key):
                self.assertTrue(
                    key in output_dict[TomoHDF5Config.ENTRIES_AND_TITLES_SECTION_DK]
                )
        # check sample pixel size
        for key in (
            TomoHDF5Config.SAMPLE_X_PIXEL_SIZE_KEYS_DK,
            TomoHDF5Config.SAMPLE_Y_PIXEL_SIZE_KEYS_DK,
            TomoHDF5Config.DETECTOR_X_PIXEL_SIZE_KEYS_DK,
            TomoHDF5Config.DETECTOR_Y_PIXEL_SIZE_KEYS_DK,
        ):
            with self.subTest(key=key):
                self.assertTrue(key in output_dict[TomoHDF5Config.KEYS_SECTION_DK])
        # translation keys
        for key in (
            TomoHDF5Config.SAMPLE_X_KEYS_DK,
            TomoHDF5Config.SAMPLE_Y_KEYS_DK,
            TomoHDF5Config.TRANSLATION_Z_KEYS_DK,
        ):
            with self.subTest(key=key):
                self.assertTrue(key in output_dict[TomoHDF5Config.KEYS_SECTION_DK])
        # others
        for key in (
            TomoHDF5Config.VALID_CAMERA_DK,
            TomoHDF5Config.ROT_ANGLE_DK,
            TomoHDF5Config.TRANSLATION_Y_KEYS_DK,
            TomoHDF5Config.DIODE_KEYS_DK,
            TomoHDF5Config.ACQUISITION_EXPO_TIME_KEYS_DK,
        ):
            with self.subTest(key=key):
                self.assertTrue(key in output_dict[TomoHDF5Config.KEYS_SECTION_DK])

    def test_from_dict(self):
        """test the `from_dict` function"""
        valid_camera_names = ("frelon", "totocam")
        alignment_titles = ("this is an alignment",)
        sample_x_keys = ("tx", "x")
        config = TomoHDF5Config.from_dict(
            {
                TomoHDF5Config.KEYS_SECTION_DK: {
                    TomoHDF5Config.VALID_CAMERA_DK: valid_camera_names,
                    TomoHDF5Config.SAMPLE_X_KEYS_DK: sample_x_keys,
                },
                TomoHDF5Config.ENTRIES_AND_TITLES_SECTION_DK: {
                    TomoHDF5Config.ALIGNMENT_TITLES_DK: alignment_titles
                },
            }
        )
        self.assertEqual(config.valid_camera_names, valid_camera_names)
        self.assertEqual(config.alignment_titles, alignment_titles)
        self.assertEqual(config.sample_x_keys, sample_x_keys)

    def test_from_dict_lowercase(self):
        """Test `from_dict` method with lower case sections"""
        ref_dict = TomoHDF5Config().to_dict()
        lower_dict = {k.lower(): v for k, v in ref_dict.items()}

        config = TomoHDF5Config.from_dict(lower_dict)
        self.assertEqual(config.to_dict(), ref_dict)

    def test_raises_errors(self):
        """
        Insure a type error is raised if an invalid type is passed to the
        HDF5Config class
        :return:
        """
        with self.assertRaises(TypeError):
            TomoHDF5Config.from_dict(
                {
                    TomoHDF5Config.ENTRIES_AND_TITLES_SECTION_DK: {
                        TomoHDF5Config.DARK_TITLES_DK: 1213
                    }
                }
            )

    def test_to_and_from_cfg_file(self):
        """
        Insure we can dump the configuration to a .cfg file and that we
        can read it back
        """
        file_path = os.path.join(self.folder, "output_file.cfg")
        input_config = TomoHDF5Config()
        input_config.to_cfg_file(file_path)
        self.assertTrue(os.path.exists(file_path))
        loaded_config = TomoHDF5Config.from_cfg_file(file_path=file_path)
        self.assertTrue(isinstance(loaded_config, TomoHDF5Config))



# coding: utf-8

from __future__ import annotations

import pint
import configparser
import logging

from typing import Iterable

from nxtomo.nxobject.nxsource import ProbeType, SourceType

from nxtomomill.io.config.optionlevel import OptionLevel, filter_options_level_items
from nxtomomill.io.config.configbase import ConfigBase, ConfigSourceSection

from nxtomomill.io.utils import (
    PathType,
    convert_str_to_bool,
    convert_str_to_tuple,
    filter_str_def,
)
from nxtomomill.settings import Tomo
from nxtomomill.utils import FileExtension
from nxtomomill.utils.pintutils import (
    VALID_CURRENT_VALUES,
    VALID_ENERGY_VALUES,
    VALID_METRIC_VALUES,
)

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)


__all__ = ["TomoEDFConfig", "generate_default_edf_config"]


class TomoEDFConfig(ConfigBase, ConfigSourceSection):
    """
    Configuration class to provide to the convert from h5 to nx
    """

    # General section keys

    GENERAL_SECTION_DK = "GENERAL_SECTION"

    INPUT_FOLDER_DK = "input_folder"

    OUTPUT_FILE_DK = "output_file"

    FILE_EXTENSION_DK = "file_extension"

    OVERWRITE_DK = "overwrite"

    DELETE_EDF_SOURCE_FILES = "delete_edf_source_file"

    OUTPUT_CHECKS = "output_checks"

    DATASET_BASENAME_DK = "dataset_basename"

    DATASET_FILE_INFO_DK = "dataset_info_file"

    LOG_LEVEL_DK = "log_level"

    TITLE_DK = "title"

    IGNORE_FILE_PATTERN_DK = "patterns_to_ignores"

    DUPLICATE_DATA_DK = "duplicate_data"

    EXTERNAL_LINK_RELATIVE_DK = "external_link_path"

    COMMENTS_GENERAL_SECTION = {
        GENERAL_SECTION_DK: "general information. \n",
        INPUT_FOLDER_DK: "Folder containing .edf files. if not provided from the configuration file must be provided from the command line",
        OUTPUT_FILE_DK: "output file name. If not provided from the configuration file must be provided from the command line",
        OVERWRITE_DK: "overwrite output files if exists without asking",
        FILE_EXTENSION_DK: "file extension. Ignored if the output file is provided and contains an extension",
        DELETE_EDF_SOURCE_FILES: "remove EDF source files once the conversion is complete. Works only if 'duplicate_data' is True (this is the case by default)",
        OUTPUT_CHECKS: "Once the conversion is done some checks can be done to check the validity of the conversion. Expected as a list of elements. Possibles tests are: 'compare-output-volume'",
        DATASET_BASENAME_DK: f"dataset file prefix. Usde to determine projections file and info file. If not provided will take the name of {INPUT_FOLDER_DK}",
        DATASET_FILE_INFO_DK: f"path to .info file containing dataset information (Energy, ScanRange, TOMO_N...). If not will deduce it from {DATASET_BASENAME_DK}",
        LOG_LEVEL_DK: 'Log level. Valid levels are "debug", "info", "warning" and "error"',
        TITLE_DK: "NXtomo title",
        IGNORE_FILE_PATTERN_DK: "some file pattern leading to ignoring the file. Like reconstructed slice files.",
        DUPLICATE_DATA_DK: "If False then will create embed all the data into a single file avoiding external link to other file. If True then the detector data will point to original edf files. In this case you must be carreful to keep relative paths valid. Warning: to read external dataset you nust be at the hdf5 file working directory. See external link resolution details: https://support.hdfgroup.org/documentation/hdf5/latest/group___h5_l.html#title5",
        EXTERNAL_LINK_RELATIVE_DK: "If 'duplicate_data' is set to False then you can specify if you want the link to original files to be 'relative' or 'absolute'",
    }

    LEVEL_GENERAL_SECTION = {
        INPUT_FOLDER_DK: OptionLevel.REQUIRED,
        OUTPUT_FILE_DK: OptionLevel.REQUIRED,
        OVERWRITE_DK: OptionLevel.REQUIRED,
        FILE_EXTENSION_DK: OptionLevel.ADVANCED,
        DELETE_EDF_SOURCE_FILES: OptionLevel.ADVANCED,
        OUTPUT_CHECKS: OptionLevel.ADVANCED,
        DATASET_BASENAME_DK: OptionLevel.REQUIRED,
        DATASET_FILE_INFO_DK: OptionLevel.REQUIRED,
        LOG_LEVEL_DK: OptionLevel.REQUIRED,
        TITLE_DK: OptionLevel.ADVANCED,
        IGNORE_FILE_PATTERN_DK: OptionLevel.REQUIRED,
        DUPLICATE_DATA_DK: OptionLevel.REQUIRED,
        EXTERNAL_LINK_RELATIVE_DK: OptionLevel.ADVANCED,
    }

    # EDF KEYS SECTION

    EDF_KEYS_SECTION_DK = "EDF_KEYS_SECTION"

    MOTOR_POSITION_KEY_DK = "motor_position_key"

    MOTOR_MNE_KEY_DK = "motor_mne_key"

    X_TRANS_KEY_DK = "x_translation_key"

    Y_TRANS_KEY_DK = "y_translation_key"

    Z_TRANS_KEY_DK = "z_translation_key"

    ROT_ANGLE_KEY_DK = "rot_angle_key"

    COMMENTS_KEYS_SECTION = {
        EDF_KEYS_SECTION_DK: "section to define EDF keys to pick from headers to deduce information like rotation angle.\n",
        MOTOR_POSITION_KEY_DK: "motor position key",
        MOTOR_MNE_KEY_DK: "key to retrieve indices of each motor in metadata",
        X_TRANS_KEY_DK: "key to be used for x translation",
        Y_TRANS_KEY_DK: "key to be used for y translation",
        Z_TRANS_KEY_DK: "key to be used for z translation",
        ROT_ANGLE_KEY_DK: "key to be used for rotation angle",
    }

    LEVEL_KEYS_SECTION = {
        MOTOR_POSITION_KEY_DK: OptionLevel.REQUIRED,
        MOTOR_MNE_KEY_DK: OptionLevel.REQUIRED,
        X_TRANS_KEY_DK: OptionLevel.REQUIRED,
        Y_TRANS_KEY_DK: OptionLevel.REQUIRED,
        Z_TRANS_KEY_DK: OptionLevel.REQUIRED,
        ROT_ANGLE_KEY_DK: OptionLevel.REQUIRED,
    }

    # DARK AND FLAT SECTION

    FLAT_DARK_SECTION_DK = "DARK_AND_FLAT_SECTION"

    DARK_NAMES_DK = "dark_names_prefix"

    FLAT_NAMES_DK = "flat_names_prefix"

    COMMENTS_DARK_FLAT_SECTION = {
        FLAT_DARK_SECTION_DK: "section to define dark and flat detection. \n",
        DARK_NAMES_DK: "prefix of dark field file(s)",
        FLAT_NAMES_DK: "prefix of flat field file(s)",
    }

    LEVEL_DARK_FLAT_SECTION = {
        DARK_NAMES_DK: OptionLevel.REQUIRED,
        FLAT_NAMES_DK: OptionLevel.REQUIRED,
    }

    # UNITS SECTION

    UNIT_SECTION_DK = "UNIT_SECTION"

    PIXEL_SIZE_EXPECTED_UNIT = "expected_unit_for_pixel_size"

    DISTANCE_EXPECTED_UNIT = "expected_unit_for_distance"

    ENERGY_EXPECTED_UNIT = "expected_unit_for_energy"

    X_TRANS_EXPECTED_UNIT = "expected_unit_for_x_translation"

    Y_TRANS_EXPECTED_UNIT = "expected_unit_for_y_translation"

    Z_TRANS_EXPECTED_UNIT = "expected_unit_for_z_translation"

    MACHINE_CURRENT_EXPECTED_UNIT = "expected_unit_for_machine_current"

    COMMENTS_UNIT_SECTION_DK = {
        UNIT_SECTION_DK: "Details units system used on SPEC side to save data. All will ne converted to NXtomo default (SI at the exception of energy-keV) \n",
        PIXEL_SIZE_EXPECTED_UNIT: f"Size used to save pixel size. Must be in of {VALID_METRIC_VALUES}",
        DISTANCE_EXPECTED_UNIT: f"Unit used by SPEC to save sample to detector distance. Must be in of {VALID_METRIC_VALUES}",
        ENERGY_EXPECTED_UNIT: f"Unit used by SPEC to save energy. Must be in of {VALID_ENERGY_VALUES}",
        X_TRANS_EXPECTED_UNIT: f"Unit used by SPEC to save x translation. Must be in of {VALID_METRIC_VALUES}",
        Y_TRANS_EXPECTED_UNIT: f"Unit used by SPEC to save y translation. Must be in of {VALID_METRIC_VALUES}",
        Z_TRANS_EXPECTED_UNIT: f"Unit used by SPEC to save z translation. Must be in of {VALID_METRIC_VALUES}",
        MACHINE_CURRENT_EXPECTED_UNIT: f"Unit used by SPEC to save machine current (also aka SRcurrent). Must be in of {VALID_CURRENT_VALUES}",
    }

    LEVEL_UNIT_SECTION = {
        PIXEL_SIZE_EXPECTED_UNIT: OptionLevel.ADVANCED,
        DISTANCE_EXPECTED_UNIT: OptionLevel.ADVANCED,
        ENERGY_EXPECTED_UNIT: OptionLevel.ADVANCED,
        X_TRANS_EXPECTED_UNIT: OptionLevel.ADVANCED,
        Y_TRANS_EXPECTED_UNIT: OptionLevel.ADVANCED,
        Z_TRANS_EXPECTED_UNIT: OptionLevel.ADVANCED,
    }

    # SAMPLE SECTION

    SAMPLE_SECTION_DK = "SAMPLE_SECTION"

    SAMPLE_NAME_DK = "sample_name"

    FORCE_ANGLE_CALCULATION = "force_angle_calculation"

    FORCE_ANGLE_CALCULATION_ENDPOINT = "angle_calculation_endpoint"

    FORCE_ANGLE_CALCULATION_REVERT_NEG_SCAN_RANGE = (
        "angle_calculation_rev_neg_scan_range"
    )

    COMMENTS_SAMPLE_SECTION_DK = {
        SAMPLE_SECTION_DK: "section dedicated to sample definition.\n",
        SAMPLE_NAME_DK: "name of the sample",
        FORCE_ANGLE_CALCULATION: "Should the rotation angle be computed from scan range and numpy.linspace or should we try to load it from .edf header.",
        FORCE_ANGLE_CALCULATION_ENDPOINT: "If rotation angles have to be calculated set numpy.linspace endpoint parameter to this value. If True then the rotation angle value of the last projection will be equal to the `ScanRange` value",
        FORCE_ANGLE_CALCULATION_REVERT_NEG_SCAN_RANGE: "Invert rotation angle values in the case of negative `ScanRange` value",
    }

    LEVEL_SAMPLE_SECTION = {
        SAMPLE_NAME_DK: OptionLevel.ADVANCED,
        FORCE_ANGLE_CALCULATION: OptionLevel.REQUIRED,
        FORCE_ANGLE_CALCULATION_ENDPOINT: OptionLevel.REQUIRED,
        FORCE_ANGLE_CALCULATION_REVERT_NEG_SCAN_RANGE: OptionLevel.ADVANCED,
    }

    # DETECTOR SECTION

    DETECTOR_SECTION_DK = "DETECTOR_SECTION"

    FIELD_OF_VIEW_DK = "field_of_view"

    COMMENTS_DETECTOR_SECTION_DK = {
        DETECTOR_SECTION_DK: "section dedicated to detector definition \n",
        FIELD_OF_VIEW_DK: "Detector field of view. Must be in `Half` or `Full`",
    }

    LEVEL_DETECTOR_SECTION = {
        FIELD_OF_VIEW_DK: OptionLevel.ADVANCED,
    }

    # create comments

    COMMENTS = COMMENTS_GENERAL_SECTION
    COMMENTS.update(COMMENTS_KEYS_SECTION)
    COMMENTS.update(COMMENTS_DARK_FLAT_SECTION)
    COMMENTS.update(COMMENTS_UNIT_SECTION_DK)
    COMMENTS.update(COMMENTS_SAMPLE_SECTION_DK)
    COMMENTS.update(ConfigSourceSection.COMMENTS_SOURCE_SECTION_DK)
    COMMENTS.update(COMMENTS_DETECTOR_SECTION_DK)

    SECTIONS_LEVEL = {
        GENERAL_SECTION_DK: OptionLevel.REQUIRED,
        EDF_KEYS_SECTION_DK: OptionLevel.REQUIRED,
        FLAT_DARK_SECTION_DK: OptionLevel.REQUIRED,
        UNIT_SECTION_DK: OptionLevel.ADVANCED,
        SAMPLE_SECTION_DK: OptionLevel.REQUIRED,
        ConfigSourceSection.SOURCE_SECTION_DK: OptionLevel.ADVANCED,
        DETECTOR_SECTION_DK: OptionLevel.ADVANCED,
    }

    def __init__(self):
        super().__init__()
        self._set_freeze(False)
        # general information
        self._input_folder = None
        self._output_file = None
        self._file_extension = FileExtension.NX
        self._overwrite = False
        self._delete_edf_source_files = False
        self._output_checks = tuple()
        self._dataset_basename = None
        self._log_level = logging.WARNING
        self._title = None
        self._ignore_file_patterns = Tomo.EDF.TO_IGNORE
        self._dataset_info_file = None
        self._duplicate_data = True
        self._external_path_type = PathType.RELATIVE

        # edf header keys
        self._motor_position_keys = Tomo.EDF.MOTOR_POS
        self._motor_mne_keys = Tomo.EDF.MOTOR_MNE
        self._x_trans_keys = Tomo.EDF.X_TRANS
        self._y_trans_keys = Tomo.EDF.Y_TRANS
        self._z_trans_keys = Tomo.EDF.Z_TRANS
        self._rot_angle_keys = Tomo.EDF.ROT_ANGLE
        self._machine_current_keys = Tomo.EDF.MACHINE_CURRENT

        # dark and flat
        self._dark_names = Tomo.EDF.DARK_NAMES
        self._flat_names = Tomo.EDF.REFS_NAMES

        # units
        self._pixel_size_unit = _ureg.micrometer
        self._distance_unit = _ureg.millimeter
        self._energy_unit = _ureg.keV
        self._x_trans_unit = _ureg.millimeter
        self._y_trans_unit = _ureg.millimeter
        self._z_trans_unit = _ureg.millimeter
        self._machine_current_unit = _ureg.milliampere

        # sample
        self._sample_name = None
        self._force_angle_calculation = True
        # there is too many EDF headers containing the rotation angle set to 0
        # when it shouldn't be. To ease usage we are 'ignoring' those keys by default...
        self._force_angle_calculation_endpoint = False
        self._force_angle_calculation_revert_neg_scan_range = True

        # source
        self._instrument_name = None
        self._source_name = "ESRF"
        self._source_type = SourceType.SYNCHROTRON_X_RAY_SOURCE
        self._source_probe = ProbeType.X_RAY

        # detector
        self._field_of_view = None

        self._set_freeze(True)

    @property
    def input_folder(self) -> str | None:
        return self._input_folder

    @input_folder.setter
    def input_folder(self, folder: str | None) -> None:
        if not isinstance(folder, (type(None), str)):
            raise TypeError(
                f"folder is expected to be None or an instance of str. Not {type(folder)}"
            )
        self._input_folder = folder

    @property
    def dataset_basename(self) -> str | None:
        return self._dataset_basename

    @dataset_basename.setter
    def dataset_basename(self, dataset_basename: str | None) -> None:
        if not isinstance(dataset_basename, (type(None), str)):
            raise TypeError(
                f"dataset_basename is expected to be None or an instance of str. Not {type(dataset_basename)}"
            )
        self._dataset_basename = dataset_basename

    @property
    def dataset_info_file(self) -> str | None:
        return self._dataset_info_file

    @dataset_info_file.setter
    def dataset_info_file(self, file_path: str | None) -> None:
        if not isinstance(file_path, (type(None), str)):
            raise TypeError(
                f"file_path is expected to be None or an instance of str. Not {type(file_path)}"
            )
        self._dataset_info_file = file_path

    @property
    def duplicate_data(self) -> bool:
        return self._duplicate_data

    @duplicate_data.setter
    def duplicate_data(self, duplicate: bool):
        if not isinstance(duplicate, bool):
            raise TypeError(
                f"duplicate is expected to be a bool and not {type(duplicate)}"
            )
        self._duplicate_data = duplicate

    @property
    def external_path_type(self) -> PathType:
        return self._external_path_type

    @external_path_type.setter
    def external_path_type(self, path_type: str | PathType):
        self._external_path_type = PathType(path_type)

    @property
    def title(self) -> str | None:
        return self._title

    @title.setter
    def title(self, title: str | None) -> None:
        if not isinstance(title, (type(None), str)):
            raise TypeError(
                f"title is expected to be None or an instance of str. Not {type(title)}"
            )
        self._title = title

    @property
    def delete_edf_source_files(self) -> bool:
        return self._delete_edf_source_files

    @delete_edf_source_files.setter
    def delete_edf_source_files(self, delete: bool) -> None:
        if not isinstance(delete, bool):
            raise TypeError("'delete' is expected to be a bool")
        self._delete_edf_source_files = delete

    @property
    def output_checks(self) -> tuple:
        return self._output_checks

    @output_checks.setter
    def output_checks(self, checks: tuple | None) -> None:
        if checks is None:
            self._output_checks = tuple()
        elif not isinstance(checks, tuple):
            raise TypeError("'checks' is expected to be None or a tuple")
        else:
            self._output_checks = checks

    @property
    def ignore_file_patterns(self) -> tuple:
        return self._ignore_file_patterns

    @ignore_file_patterns.setter
    def ignore_file_patterns(self, patterns: tuple | list | None):
        if not isinstance(patterns, (type(None), tuple, list)):
            raise TypeError("patterns is expected to be a tuple or a list")
        if patterns is None:
            self._ignore_file_patterns = tuple()
        else:
            for elmt in patterns:
                if not isinstance(elmt, str):
                    raise TypeError("patterns elmts are expected to be str")
            self._ignore_file_patterns = tuple(patterns)

    @property
    def motor_position_keys(self) -> tuple:
        return self._motor_position_keys

    @motor_position_keys.setter
    def motor_position_keys(self, keys: tuple | list | None) -> None:
        if not isinstance(keys, (type(None), tuple, list)):
            raise TypeError(
                "keys is expected to be None or an instance of list or tuple"
            )
        if keys is None:
            self._motor_position_keys = tuple()
        else:
            for elmt in keys:
                if not isinstance(elmt, str):
                    raise TypeError("keys elmts are expected to be str")
            self._motor_position_keys = tuple(keys)

    @property
    def motor_mne_keys(self) -> tuple:
        return self._motor_mne_keys

    @motor_mne_keys.setter
    def motor_mne_keys(self, keys: tuple | list | None) -> None:
        if not isinstance(keys, (type(None), tuple, list)):
            raise TypeError(
                "keys is expected to be None or an instance of list or tuple"
            )
        if keys is None:
            self._motor_mne_keys = tuple()
        else:
            for elmt in keys:
                if not isinstance(elmt, str):
                    raise TypeError("keys elmts are expected to be str")
            self._motor_mne_keys = tuple(keys)

    @property
    def dark_names(self) -> tuple:
        return self._dark_names

    @dark_names.setter
    def dark_names(self, names: tuple | list | None) -> None:
        if not isinstance(names, (type(None), tuple, list)):
            raise TypeError("names is expected to be a tuple or a list")
        if names is None:
            self._dark_names = tuple()
        else:
            for elmt in names:
                if not isinstance(elmt, str):
                    raise TypeError("names elmts are expected to be str")
            self._dark_names = tuple(names)

    @property
    def flat_names(self) -> tuple:
        return self._flat_names

    @flat_names.setter
    def flat_names(self, names: tuple | list | None) -> None:
        if not isinstance(names, (type(None), tuple, list)):
            raise TypeError("names is expected to be a tuple or a list")
        if names is None:
            self._flat_names = tuple()
        else:
            for elmt in names:
                if not isinstance(elmt, str):
                    raise TypeError("names elmts are expected to be str")
            self._flat_names = tuple(names)

    @property
    def pixel_size_unit(self) -> pint.Unit:
        return self._pixel_size_unit

    @pixel_size_unit.setter
    def pixel_size_unit(self, unit: pint.Unit | str) -> None:
        if not isinstance(unit, (pint.Unit, str)):
            raise TypeError("unit is expected to be an instance of pint.Unit")
        if isinstance(unit, str):
            unit = pint.Unit(unit)
        self._pixel_size_unit = unit

    @property
    def distance_unit(self) -> pint.Unit:
        return self._distance_unit

    @distance_unit.setter
    def distance_unit(self, unit: pint.Unit | str) -> None:
        if not isinstance(unit, (pint.Unit, str)):
            raise TypeError("unit is expected to be an instance of pint.Unit")
        if isinstance(unit, str):
            unit = pint.Unit(unit)
        self._distance_unit = unit

    @property
    def energy_unit(self) -> pint.Unit:
        return self._energy_unit

    @energy_unit.setter
    def energy_unit(self, unit: pint.Unit | str) -> None:
        if not isinstance(unit, (pint.Unit, str)):
            raise TypeError("unit is expected to be an instance of pint.Unit")
        if isinstance(unit, str):
            unit = pint.Unit(unit)
        self._energy_unit = unit

    @property
    def x_trans_keys(self):
        return self._x_trans_keys

    @x_trans_keys.setter
    def x_trans_keys(self, keys: Iterable) -> None:
        if not isinstance(keys, Iterable) or isinstance(keys, str):
            raise TypeError("'keys' should be an Iterable")

        for elmt in keys:
            if not isinstance(elmt, str):
                raise TypeError("keys elmts are expected to be str")
        self._x_trans_keys = keys

    @property
    def x_trans_unit(self) -> pint.Unit:
        return self._x_trans_unit

    @x_trans_unit.setter
    def x_trans_unit(self, unit: pint.Unit | str) -> None:
        if not isinstance(unit, (pint.Unit, str)):
            raise TypeError("unit is expected to be an instance of pint.unit")
        if isinstance(unit, str):
            unit = pint.Unit(unit)
        self._x_trans_unit = unit

    @property
    def y_trans_keys(self):
        return self._y_trans_keys

    @y_trans_keys.setter
    def y_trans_keys(self, keys: Iterable) -> None:
        if not isinstance(keys, Iterable) or isinstance(keys, str):
            raise TypeError("'keys' should be an Iterable")

        for elmt in keys:
            if not isinstance(elmt, str):
                raise TypeError("keys elmts are expected to be str")
        self._y_trans_keys = keys

    @property
    def y_trans_unit(self) -> pint.Unit:
        return self._y_trans_unit

    @y_trans_unit.setter
    def y_trans_unit(self, unit: pint.Unit | str) -> None:
        if not isinstance(unit, (pint.Unit, str)):
            raise TypeError("unit is expected to be an instance of pint.Unit")
        if isinstance(unit, str):
            unit = pint.Unit(unit)
        self._y_trans_unit = unit

    @property
    def z_trans_keys(self):
        return self._z_trans_keys

    @z_trans_keys.setter
    def z_trans_keys(self, keys: Iterable) -> None:
        if not isinstance(keys, Iterable) or isinstance(keys, str):
            raise TypeError("'keys' should be an Iterable")

        for elmt in keys:
            if not isinstance(elmt, str):
                raise TypeError("keys elmts are expected to be str")
        self._z_trans_keys = keys

    @property
    def z_trans_unit(self) -> pint.Unit:
        return self._z_trans_unit

    @z_trans_unit.setter
    def z_trans_unit(self, unit: pint.Unit | str) -> None:
        if not isinstance(unit, (pint.Unit, str)):
            raise TypeError("unit is expected to be an instance of pint.Unit")
        if isinstance(unit, str):
            unit = pint.Unit(unit)
        self._z_trans_unit = unit

    @property
    def machine_current_unit(self) -> pint.Unit:
        return self._machine_current_unit

    @machine_current_unit.setter
    def machine_current_unit(self, unit):
        if not isinstance(unit, (str, pint.Unit)):
            raise TypeError("unit is expected to be an instance of pint.Unit")
        if isinstance(unit, str):
            unit = pint.Unit(unit)
        self._machine_current_unit = unit

    @property
    def sample_name(self) -> str | None:
        return self._sample_name

    @sample_name.setter
    def sample_name(self, name: str | None) -> None:
        if not isinstance(name, (type(None), str)):
            raise TypeError("name is expected to be None or an instance of str")
        self._sample_name = name

    @property
    def force_angle_calculation(self) -> bool:
        return self._force_angle_calculation

    @force_angle_calculation.setter
    def force_angle_calculation(self, force: bool):
        if not isinstance(force, bool):
            raise TypeError(
                f"force is expected to be an instance of bool. Not {type(force)}  - {force}"
            )
        self._force_angle_calculation = force

    @property
    def force_angle_calculation_endpoint(self) -> bool:
        return self._force_angle_calculation_endpoint

    @force_angle_calculation_endpoint.setter
    def force_angle_calculation_endpoint(self, endpoint: bool) -> None:
        if not isinstance(endpoint, bool):
            raise TypeError(
                f"endpoint is expected to be an instance of bool. Not {type(endpoint)}"
            )
        self._force_angle_calculation_endpoint = endpoint

    @property
    def angle_calculation_rev_neg_scan_range(self) -> bool:
        return self._force_angle_calculation_revert_neg_scan_range

    @angle_calculation_rev_neg_scan_range.setter
    def angle_calculation_rev_neg_scan_range(self, revert: bool):
        if not isinstance(revert, bool):
            raise TypeError(
                f"revert is expected to be an instance of bool. Not {type(revert)}"
            )
        self._force_angle_calculation_revert_neg_scan_range = revert

    @property
    def instrument_name(self) -> str | None:
        return self._instrument_name

    @instrument_name.setter
    def instrument_name(self, name: str | None):
        if not isinstance(name, (type(None), str)):
            raise TypeError("name is expected to be None or an instance of str")
        self._instrument_name = name

    @property
    def source_name(self) -> str | None:
        return self._source_name

    @source_name.setter
    def source_name(self, name: str | None) -> None:
        if not isinstance(name, (type(None), str)):
            raise TypeError("name is expected to be None or an instance of str")
        self._source_name = name

    @property
    def source_type(self) -> SourceType | None:
        return self._source_type

    @source_type.setter
    def source_type(self, source_type: SourceType | str | None):
        if not isinstance(source_type, (type(None), str, SourceType)):
            raise TypeError(
                "source_type is expected to be None or an instance of SourceType or str"
            )
        if source_type is None:
            self._source_type = None
        else:
            self._source_type = SourceType(source_type)

    @property
    def source_probe(self) -> ProbeType | None:
        return self._source_probe

    @source_probe.setter
    def source_probe(self, source_probe: ProbeType | str | None):
        if not isinstance(source_probe, (type(None), str, ProbeType)):
            raise TypeError(
                "source_probe is expected to be None or an instance of ProbeType or str"
            )
        if source_probe is None:
            self._source_probe = None
        else:
            self._source_probe = ProbeType(source_probe)

    def to_dict(self, level="advanced") -> dict:
        """convert the configuration to a dictionary"""
        level = OptionLevel(level)
        sections_callback = {
            self.GENERAL_SECTION_DK: self._general_section_to_dict,
            self.EDF_KEYS_SECTION_DK: self._edf_keys_section_to_dict,
            self.FLAT_DARK_SECTION_DK: self._flat_keys_section_to_dict,
            self.UNIT_SECTION_DK: self._unit_section_to_dict,
            self.SAMPLE_SECTION_DK: self._sample_section_to_dict,
            self.SOURCE_SECTION_DK: self._source_section_to_dict,
            self.DETECTOR_SECTION_DK: self._detector_section_to_dict,
        }
        res = {}
        for section, callback in sections_callback.items():
            if (
                level == OptionLevel.ADVANCED
                or TomoEDFConfig.SECTIONS_LEVEL[section] == OptionLevel.REQUIRED
            ):
                res[section] = callback(level=level)
        return res

    def _general_section_to_dict(self, level) -> dict:
        res = {
            self.INPUT_FOLDER_DK: (
                self.input_folder if self.input_folder is not None else ""
            ),
            self.OUTPUT_FILE_DK: (
                self.output_file if self.output_file is not None else ""
            ),
            self.OVERWRITE_DK: self.overwrite,
            self.DELETE_EDF_SOURCE_FILES: self.delete_edf_source_files,
            self.OUTPUT_CHECKS: self.output_checks,
            self.FILE_EXTENSION_DK: self.file_extension.value,
            self.DATASET_BASENAME_DK: (
                self.dataset_basename if self.dataset_basename is not None else ""
            ),
            self.DATASET_FILE_INFO_DK: (
                self.dataset_info_file if self.dataset_info_file is not None else ""
            ),
            self.LOG_LEVEL_DK: logging.getLevelName(self.log_level).lower(),
            self.TITLE_DK: self.title if self.title is not None else "",
            self.IGNORE_FILE_PATTERN_DK: (
                self.ignore_file_patterns
                if self.ignore_file_patterns != tuple()
                else ""
            ),
            self.DUPLICATE_DATA_DK: self.duplicate_data,
            self.EXTERNAL_LINK_RELATIVE_DK: self.external_path_type.value,
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=TomoEDFConfig.LEVEL_GENERAL_SECTION
        )

    def _edf_keys_section_to_dict(self, level) -> dict:
        res = {
            self.MOTOR_POSITION_KEY_DK: (
                self.motor_position_keys if self.motor_position_keys != tuple() else ""
            ),
            self.MOTOR_MNE_KEY_DK: (
                self.motor_mne_keys if self.motor_mne_keys != tuple() else ""
            ),
            self.ROT_ANGLE_KEY_DK: (
                self.rotation_angle_keys if self.rotation_angle_keys != tuple() else ""
            ),
            self.X_TRANS_KEY_DK: (
                self.x_trans_keys if self.x_trans_keys != tuple() else ""
            ),
            self.Y_TRANS_KEY_DK: (
                self.y_trans_keys if self.y_trans_keys != tuple() else ""
            ),
            self.Z_TRANS_KEY_DK: (
                self.z_trans_keys if self.z_trans_keys != tuple() else ""
            ),
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=TomoEDFConfig.LEVEL_KEYS_SECTION
        )

    def _flat_keys_section_to_dict(self, level) -> dict:
        res = {
            self.DARK_NAMES_DK: self.dark_names if self.dark_names != tuple() else "",
            self.FLAT_NAMES_DK: self.flat_names if self.dark_names != tuple() else "",
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=TomoEDFConfig.LEVEL_DARK_FLAT_SECTION
        )

    def _unit_section_to_dict(self, level) -> dict:
        res = {
            self.PIXEL_SIZE_EXPECTED_UNIT: str(self.pixel_size_unit),
            self.DISTANCE_EXPECTED_UNIT: str(self.distance_unit),
            self.ENERGY_EXPECTED_UNIT: str(self.energy_unit),
            self.X_TRANS_EXPECTED_UNIT: str(self.x_trans_unit),
            self.Y_TRANS_EXPECTED_UNIT: str(self.y_trans_unit),
            self.Z_TRANS_EXPECTED_UNIT: str(self.z_trans_unit),
            self.MACHINE_CURRENT_EXPECTED_UNIT: str(self.machine_current_unit),
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=TomoEDFConfig.LEVEL_UNIT_SECTION
        )

    def _sample_section_to_dict(self, level) -> dict:
        res = {
            self.SAMPLE_NAME_DK: (
                self.sample_name if self.sample_name is not None else ""
            ),
            self.FORCE_ANGLE_CALCULATION: self.force_angle_calculation,
            self.FORCE_ANGLE_CALCULATION_ENDPOINT: self.force_angle_calculation_endpoint,
            self.FORCE_ANGLE_CALCULATION_REVERT_NEG_SCAN_RANGE: self.angle_calculation_rev_neg_scan_range,
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=TomoEDFConfig.LEVEL_SAMPLE_SECTION
        )

    def _source_section_to_dict(self, level) -> dict:
        res = {
            self.INSTRUMENT_NAME_DK: self.instrument_name or "",
            self.SOURCE_NAME_DK: self.source_name or "",
            self.SOURCE_TYPE_DK: (
                self.source_type.value if self.source_type is not None else ""
            ),
            self.SOURCE_PROBE_DK: (
                self.source_probe.value if self.source_probe is not None else ""
            ),
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=TomoEDFConfig.LEVEL_SOURCE_SECTION
        )

    def _detector_section_to_dict(self, level) -> dict:
        res = {
            self.FIELD_OF_VIEW_DK: (
                self.field_of_view.value if self.field_of_view is not None else ""
            ),
        }
        return filter_options_level_items(
            dict_=res, level=level, level_ref=TomoEDFConfig.LEVEL_DETECTOR_SECTION
        )

    @staticmethod
    def from_dict(dict_: dict):
        r"""
        Create a HDF5Config object and set it from values contained in the
        dictionary
        :param dict\_: settings dictionary
        :return: HDF5Config
        """
        config = TomoEDFConfig()
        config.load_from_dict(dict_)
        return config

    def load_from_dict(self, dict_: dict) -> None:
        """Load the configuration from a dictionary"""
        sections_loaders = {
            TomoEDFConfig.GENERAL_SECTION_DK: self.load_general_section,
            TomoEDFConfig.EDF_KEYS_SECTION_DK: self.load_keys_section,
            TomoEDFConfig.FLAT_DARK_SECTION_DK: self.load_flat_dark_section,
            TomoEDFConfig.UNIT_SECTION_DK: self.load_unit_section,
            TomoEDFConfig.SAMPLE_SECTION_DK: self.load_sample_section,
            TomoEDFConfig.SOURCE_SECTION_DK: self.load_source_section,
            TomoEDFConfig.DETECTOR_SECTION_DK: self.load_detector_section,
        }
        for section_key, loaded_func in sections_loaders.items():
            if section_key in dict_:
                loaded_func(dict_[section_key])
            else:
                _logger.info(
                    f"No {section_key} section found. Will take default values"
                )

    def load_general_section(self, dict_: dict) -> None:
        self.input_folder = dict_.get(TomoEDFConfig.INPUT_FOLDER_DK, None)
        self.output_file = dict_.get(TomoEDFConfig.OUTPUT_FILE_DK, None)
        overwrite = dict_.get(TomoEDFConfig.OVERWRITE_DK, None)
        if overwrite is not None:
            self.overwrite = convert_str_to_bool(overwrite)

        delete_edf_source_files = dict_.get(TomoEDFConfig.DELETE_EDF_SOURCE_FILES, None)
        if delete_edf_source_files is not None:
            self.delete_edf_source_files = convert_str_to_bool(delete_edf_source_files)

        output_checks = dict_.get(TomoEDFConfig.OUTPUT_CHECKS, None)
        if output_checks is not None:
            self.output_checks = convert_str_to_tuple(output_checks)

        file_extension = dict_.get(TomoEDFConfig.FILE_EXTENSION_DK, None)
        if file_extension not in (None, ""):
            self.file_extension = filter_str_def(file_extension)

        dataset_basename = dict_.get(TomoEDFConfig.DATASET_BASENAME_DK, None)
        if dataset_basename is not None:
            if dataset_basename == "":
                dataset_basename = None
            self.dataset_basename = dataset_basename

        dataset_info_file = dict_.get(TomoEDFConfig.DATASET_FILE_INFO_DK, None)
        if dataset_info_file is not None:
            if dataset_info_file == "":
                dataset_info_file = None
            self.dataset_info_file = dataset_info_file

        log_level = dict_.get(TomoEDFConfig.LOG_LEVEL_DK, None)
        if log_level is not None:
            self.log_level = log_level
        self.title = dict_.get(TomoEDFConfig.TITLE_DK)
        ignore_file_patterns = dict_.get(TomoEDFConfig.IGNORE_FILE_PATTERN_DK, None)
        if ignore_file_patterns is not None:
            if ignore_file_patterns == "":
                ignore_file_patterns = tuple()
            else:
                ignore_file_patterns = convert_str_to_tuple(ignore_file_patterns)
            self.ignore_file_patterns = ignore_file_patterns
        duplicate_data = dict_.get(TomoEDFConfig.DUPLICATE_DATA_DK, None)
        if duplicate_data is not None:
            self.duplicate_data = convert_str_to_bool(duplicate_data)
        external_path_type = dict_.get(TomoEDFConfig.EXTERNAL_LINK_RELATIVE_DK, None)
        if external_path_type is not None:
            self.external_path_type = external_path_type

    def load_keys_section(self, dict_: dict) -> None:
        motor_position_keys = dict_.get(TomoEDFConfig.MOTOR_POSITION_KEY_DK, None)
        if motor_position_keys is not None:
            if motor_position_keys == "":
                motor_position_keys = tuple()
            else:
                motor_position_keys = convert_str_to_tuple(motor_position_keys)
            self.motor_position_keys = motor_position_keys

        motor_mne_keys = dict_.get(TomoEDFConfig.MOTOR_MNE_KEY_DK, None)
        if motor_mne_keys is not None:
            if motor_mne_keys == "":
                motor_mne_keys = tuple()
            else:
                motor_mne_keys = convert_str_to_tuple(motor_mne_keys)
            self.motor_mne_keys = motor_mne_keys

        rotation_angle_keys = dict_.get(TomoEDFConfig.ROT_ANGLE_KEY_DK, None)
        if rotation_angle_keys is not None:
            if rotation_angle_keys == "":
                rotation_angle_keys = tuple()
            else:
                rotation_angle_keys = convert_str_to_tuple(rotation_angle_keys)
            self.rotation_angle_keys = rotation_angle_keys

        x_trans_keys = dict_.get(TomoEDFConfig.X_TRANS_KEY_DK, None)
        if x_trans_keys is not None:
            if x_trans_keys == "":
                x_trans_keys = tuple()
            else:
                x_trans_keys = convert_str_to_tuple(x_trans_keys)
            self.x_trans_keys = x_trans_keys

        y_trans_keys = dict_.get(TomoEDFConfig.Y_TRANS_KEY_DK, None)
        if y_trans_keys is not None:
            if y_trans_keys == "":
                y_trans_keys = tuple()
            else:
                y_trans_keys = convert_str_to_tuple(y_trans_keys)
            self.y_trans_keys = y_trans_keys

        z_trans_keys = dict_.get(TomoEDFConfig.Z_TRANS_KEY_DK, None)
        if z_trans_keys is not None:
            if z_trans_keys == "":
                z_trans_keys = tuple()
            else:
                z_trans_keys = convert_str_to_tuple(z_trans_keys)
            self.z_trans_keys = z_trans_keys

    def load_flat_dark_section(self, dict_: dict) -> None:
        dark_names = dict_.get(TomoEDFConfig.DARK_NAMES_DK, None)
        if dark_names is not None:
            if dark_names == "":
                dark_names = tuple()
            else:
                dark_names = convert_str_to_tuple(dark_names)
            self.dark_names = dark_names

        flat_names = dict_.get(TomoEDFConfig.FLAT_NAMES_DK, None)
        if flat_names is not None:
            if flat_names == "":
                flat_names = tuple()
            else:
                flat_names = convert_str_to_tuple(flat_names)
            self.flat_names = flat_names

    def load_unit_section(self, dict_: dict) -> None:
        if TomoEDFConfig.PIXEL_SIZE_EXPECTED_UNIT in dict_:
            self.pixel_size_unit = pint.Unit(
                dict_.get(TomoEDFConfig.PIXEL_SIZE_EXPECTED_UNIT)
            )
        if TomoEDFConfig.DISTANCE_EXPECTED_UNIT in dict_:
            self.distance_unit = pint.Unit(
                dict_.get(TomoEDFConfig.DISTANCE_EXPECTED_UNIT)
            )
        if TomoEDFConfig.ENERGY_EXPECTED_UNIT in dict_:
            self.energy_unit = pint.Unit(dict_.get(TomoEDFConfig.ENERGY_EXPECTED_UNIT))
        if TomoEDFConfig.X_TRANS_EXPECTED_UNIT in dict_:
            self.x_trans_unit = pint.Unit(
                dict_.get(TomoEDFConfig.X_TRANS_EXPECTED_UNIT)
            )
        if TomoEDFConfig.Y_TRANS_EXPECTED_UNIT in dict_:
            self.y_trans_unit = pint.Unit(
                dict_.get(TomoEDFConfig.Y_TRANS_EXPECTED_UNIT)
            )
        if TomoEDFConfig.Z_TRANS_EXPECTED_UNIT in dict_:
            self.z_trans_unit = pint.Unit(
                dict_.get(TomoEDFConfig.Z_TRANS_EXPECTED_UNIT)
            )
        if TomoEDFConfig.MACHINE_CURRENT_EXPECTED_UNIT in dict_:
            self.machine_current_unit = pint.Unit(
                dict_.get(TomoEDFConfig.MACHINE_CURRENT_EXPECTED_UNIT)
            )

    def load_sample_section(self, dict_: dict) -> None:
        if TomoEDFConfig.SAMPLE_NAME_DK in dict_:
            self.sample_name = dict_.get(TomoEDFConfig.SAMPLE_NAME_DK)

        force_angle_calculation = dict_.get(TomoEDFConfig.FORCE_ANGLE_CALCULATION, None)
        if force_angle_calculation is not None:
            self.force_angle_calculation = convert_str_to_bool(force_angle_calculation)

        force_angle_calculation_endpoint = dict_.get(
            TomoEDFConfig.FORCE_ANGLE_CALCULATION_ENDPOINT, None
        )
        if force_angle_calculation_endpoint is not None:
            self.force_angle_calculation_endpoint = convert_str_to_bool(
                force_angle_calculation_endpoint
            )

        angle_calculation_rev_neg_scan_range = dict_.get(
            TomoEDFConfig.FORCE_ANGLE_CALCULATION_REVERT_NEG_SCAN_RANGE, None
        )
        if angle_calculation_rev_neg_scan_range is not None:
            self.angle_calculation_rev_neg_scan_range = convert_str_to_bool(
                angle_calculation_rev_neg_scan_range
            )

    def load_detector_section(self, dict_: dict) -> None:
        field_of_view = dict_.get(TomoEDFConfig.FIELD_OF_VIEW_DK, None)
        if field_of_view is not None:
            if field_of_view == "":
                field_of_view = None
            self.field_of_view = field_of_view

    def to_cfg_file(self, file_path: str):
        # TODO: add some generic information like:provided order of the tuple
        # will be the effective one. You can provide a key from it names if
        # it is contained in the positioners group
        # maybe split in sub section ?
        self.dict_to_cfg(file_path=file_path, dict_=self.to_dict())

    @staticmethod
    def dict_to_cfg(file_path, dict_):
        """ """
        return ConfigBase._dict_to_cfg(
            file_path=file_path,
            dict_=dict_,
            comments_fct=TomoEDFConfig.get_comments,
            logger=_logger,
        )

    @staticmethod
    def from_cfg_file(file_path: str, encoding=None):
        assert file_path is not None, "file_path should not be None"
        config_parser = configparser.ConfigParser(allow_no_value=True)
        config_parser.read(file_path, encoding=encoding)
        return TomoEDFConfig.from_dict(config_parser)

    @staticmethod
    def get_comments(key):
        return TomoEDFConfig.COMMENTS[key]


def generate_default_edf_config(level: str = "required") -> dict:
    """generate a default configuration for converting spec-edf to NXtomo"""
    return TomoEDFConfig().to_dict(level=level)



# coding: utf-8

"""
Utils related to bliss-HDF5
"""

from __future__ import annotations
from typing import Iterable

import h5py
import numpy
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as open_hdf5

from nxtomo.nxobject.nxdetector import ImageKey
from nxtomomill.io.acquisitionstep import AcquisitionStep
from nxtomomill.io.config import TomoHDF5Config

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass
import logging

_logger = logging.getLogger(__name__)


__all__ = [
    "has_valid_detector",
    "get_entry_type",
    "get_nx_detectors",
    "guess_nx_detector",
    "deduce_machine_current",
    "split_timestamps",
]


def has_valid_detector(node, detectors_names):
    """
    :return True if the node looks like a valid nx detector
    """
    for key in node.keys():
        if (
            "NX_class" in node[key].attrs
            and node[key].attrs["NX_class"] == "NXdetector"
        ):
            if detectors_names is None or key in detectors_names:
                return True
    return False


def get_entry_type(
    url: DataUrl, configuration: TomoHDF5Config
) -> AcquisitionStep | None:
    """
    :param url: bliss scan url to type
    :return: return the step of the acquisition or None if cannot find it.
    """
    if not isinstance(url, DataUrl):
        raise TypeError(f"DataUrl is expected. Not {type(url)}")
    if url.data_slice() is not None:
        raise ValueError(
            "url expect to provide a link to bliss scan. Slice " "are not handled"
        )

    def _get_entry_type_from_title(entry: h5py.Group):
        """
        try to determine the entry type from the title
        """
        try:
            title = h5py_read_dataset(entry["title"])
        except Exception:
            _logger.error(f"fail to find title for {entry.name}, skip this group")
            return None
        else:
            init_titles = list(configuration.init_titles)
            init_titles.extend(configuration.zserie_init_titles)
            init_titles.extend(configuration.multitomo_init_titles)
            init_titles.extend(configuration.back_and_forth_init_titles)

            step_titles = {
                AcquisitionStep.INITIALIZATION: init_titles,
                AcquisitionStep.DARK: configuration.dark_titles,
                AcquisitionStep.FLAT: configuration.flat_titles,
                AcquisitionStep.PROJECTION: configuration.projections_titles,
                AcquisitionStep.ALIGNMENT: configuration.alignment_titles,
            }

            for step, titles in step_titles.items():
                for title_start in titles:
                    if title.startswith(title_start):
                        return step
            return None

    def _get_entry_type_from_technique_img_key(entry: h5py.Group):
        """
        try to determine entry type from the scan/technique sub groups.
        If this is a flat then we expect to have a "flat" group. If this is a set of projection we expect to have a "proj" group.
        For now alignment / return are unfilled
        """
        group_technique = entry.get("technique", dict())
        if "image_key" not in group_technique:
            return None

        image_key = h5py_read_dataset(group_technique["image_key"])
        if image_key is None:
            return None
        else:
            try:
                image_key = ImageKey(image_key)
            except ValueError:
                _logger.error(f"unrecognized image key: '{image_key}'")
                return None
            else:
                connections = {
                    ImageKey.DARK_FIELD: AcquisitionStep.DARK,
                    ImageKey.FLAT_FIELD: AcquisitionStep.FLAT,
                    ImageKey.PROJECTION: AcquisitionStep.PROJECTION,
                    ImageKey.ALIGNMENT: AcquisitionStep.ALIGNMENT,
                }
                return connections.get(image_key, None)

    with open_hdf5(url.file_path()) as h5f:
        if url.data_path() not in h5f:
            raise ValueError(f"Provided path does not exists: {url}")
        entry = h5f[url.data_path()]
        if not isinstance(entry, h5py.Group):
            raise ValueError(
                f"Expected path is not related to a h5py.Group ({entry}) when expect to target a bliss entry."
            )
        return _get_entry_type_from_technique_img_key(
            entry
        ) or _get_entry_type_from_title(entry)


def get_nx_detectors(node: h5py.Group) -> tuple:
    """

    :param node: node to inspect
    :return: tuple of NXdetector (h5py.Group) contained in `node`
             (expected to be the `instrument` group)
    """
    if not isinstance(node, h5py.Group):
        raise TypeError("node should be an instance of h5py.Group")
    nx_detectors = []
    for _, subnode in node.items():
        if isinstance(subnode, h5py.Group) and "NX_class" in subnode.attrs:
            if subnode.attrs["NX_class"] == "NXdetector":
                if "data" in subnode and hasattr(subnode["data"], "ndim"):
                    if subnode["data"].ndim == 3:
                        nx_detectors.append(subnode)
    nx_detectors = sorted(nx_detectors, key=lambda det: det.name)
    return tuple(nx_detectors)


def guess_nx_detector(node: h5py.Group) -> tuple:
    """
    Try to guess what can be an nx_detector without using the "NXdetector"
    NX_class attribute. Expect to find a 3D dataset named 'data' under
    a subnode
    """
    if not isinstance(node, h5py.Group):
        raise TypeError("node should be an instance of h5py.Group")
    nx_detectors = []
    for _, subnode in node.items():
        if isinstance(subnode, h5py.Group) and "data" in subnode:
            if isinstance(subnode["data"], h5py.Dataset) and subnode["data"].ndim == 3:
                nx_detectors.append(subnode)

    nx_detectors = sorted(nx_detectors, key=lambda det: det.name)
    return tuple(nx_detectors)


def deduce_machine_current(timestamps: tuple, known_machine_current: dict) -> tuple:
    """
    :param known_machine_current: keys are timestamp. Value is machine current
    :param timestamps: timestamp for which we want to get machine current
    """
    if not isinstance(known_machine_current, dict):
        raise TypeError("known_machine_current is expected to be a dict")
    for elmt in timestamps:
        if not isinstance(elmt, numpy.datetime64):
            raise TypeError(
                f"Elements of timestamps are expected to be {numpy.datetime64} and not {type(elmt)}"
            )
    if len(known_machine_current) == 0:
        raise ValueError("known_machine_current should contain at least one element")
    for key, value in known_machine_current.items():
        if not isinstance(key, numpy.datetime64):
            raise TypeError(
                f"known_machine_current keys are expected to be instances of {numpy.datetime64} and not {type(key)}"
            )
        if not isinstance(value, (float, numpy.number)):
            raise TypeError(
                "known_machine_current values are expected to be instances of float"
            )

    # 1. Order **known** machine current by time stamps (key)
    known_machine_current = dict(sorted(known_machine_current.items()))
    known_timestamps = numpy.fromiter(
        known_machine_current.keys(),
        dtype="datetime64[ns]",
        count=len(known_machine_current),
    )
    known_machine_current_values = numpy.fromiter(
        known_machine_current.values(),
        dtype="float64",
        count=len(known_machine_current),
    )

    # 2. Sort the supplied timestamps
    timestamp_input_ordering = numpy.argsort(numpy.array(timestamps))
    timestamps_sorted = numpy.take_along_axis(
        numpy.array(timestamps), indices=timestamp_input_ordering, axis=0
    )

    timestamps_sorted = numpy.sort(numpy.array(timestamps, dtype="datetime64[ns]"))

    # 3. Convert to float for numpy.interp
    known_timestamps_float = known_timestamps.astype("float64")
    timestamps_float = timestamps_sorted.astype("float64")

    # 4. Interpolate the values
    interpolated_values = numpy.interp(
        timestamps_float, known_timestamps_float, known_machine_current_values
    )

    # 5. Reorder interpolated values to match the order of original timestamps
    ordered_interpolated_values = numpy.zeros_like(interpolated_values)
    for i, o_pos in enumerate(timestamp_input_ordering):
        ordered_interpolated_values[o_pos] = interpolated_values[i]
    return tuple(ordered_interpolated_values)


def split_timestamps(my_array: Iterable, n_part: int):
    """
    split given array into n_part (as equal as possible)
    :param Iterable my_array:
    """
    array_size = len(my_array)
    if array_size < n_part:
        yield my_array
    else:
        start = 0
        for _ in range(n_part):
            end = max(start + int(array_size / n_part) + 1, array_size)
            yield my_array[start:end]
            start = end


def group_series(acquisition, list_of_series: list) -> list:
    """
    :param ZSeriesBaseAcquisition acquisition:
    z-series version 2 and 3 are all defined in a separate sequence.
    So we need to aggregate for post processing based on there names.
    post-processing can be dark / flat copy to others NXtomo
    """
    for series in list_of_series:
        if series[0].is_part_of_same_series(acquisition):
            series.append(acquisition)
            return list_of_series
    list_of_series.append(
        [
            acquisition,
        ]
    )
    return list_of_series



# coding: utf-8

import os
import numpy
import pytest
import h5py

from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill.io.config import TomoHDF5Config
from nxtomomill.io.acquisitionstep import AcquisitionStep
from nxtomomill.converter.hdf5.acquisition.utils import (
    deduce_machine_current,
    split_timestamps,
    get_entry_type,
    group_series,
)
from nxtomomill.converter.hdf5.acquisition.zseriesacquisition import (
    ZSeriesBaseAcquisition,
)
from nxtomomill.utils.utils import str_datetime_to_numpy_datetime64


def test_deduce_machine_current():
    """
    Test `deduce_current` function. Base function to compute current for each frame according to it's timestamp
    """

    current_datetimes = {
        "2022-01-15T21:07:58.360095+02:00": 1.1,
        "2022-04-15T21:07:58.360095+02:00": 121.1,
        "2022-04-15T21:09:58.360095+02:00": 123.3,
        "2022-04-15T21:11:58.360095+02:00": 523.3,
        "2022-12-15T21:07:58.360095+02:00": 1000.3,
    }
    with pytest.raises(ValueError):
        deduce_machine_current(tuple(), {})
    with pytest.raises(TypeError):
        deduce_machine_current(12, 2)
    with pytest.raises(TypeError):
        deduce_machine_current(
            [
                12,
            ],
            2,
        )
    with pytest.raises(TypeError):
        deduce_machine_current(
            [
                2,
            ],
            current_datetimes,
        )

    converted_currents = {}
    for elec_cur_datetime_str, elect_cur in current_datetimes.items():
        datetime_as_datetime = str_datetime_to_numpy_datetime64(elec_cur_datetime_str)
        converted_currents[datetime_as_datetime] = elect_cur

    # check exacts values, left and right bounds
    assert deduce_machine_current(
        (str_datetime_to_numpy_datetime64("2022-01-15T21:07:58.360095+02:00"),),
        converted_currents,
    ) == (1.1,)
    assert deduce_machine_current(
        (str_datetime_to_numpy_datetime64("2022-01-14T21:07:58.360095+02:00"),),
        converted_currents,
    ) == (1.1,)
    assert deduce_machine_current(
        (str_datetime_to_numpy_datetime64("2022-12-15T21:07:58.360095+02:00"),),
        converted_currents,
    ) == (1000.3,)
    assert deduce_machine_current(
        (str_datetime_to_numpy_datetime64("2022-12-16T21:07:58.360095+02:00"),),
        converted_currents,
    ) == (1000.3,)
    # check interpolated values
    numpy.testing.assert_almost_equal(
        deduce_machine_current(
            (str_datetime_to_numpy_datetime64("2022-04-15T21:08:58.360095+02:00"),),
            converted_currents,
        )[0],
        (122.2,),
    )
    numpy.testing.assert_almost_equal(
        deduce_machine_current(
            (str_datetime_to_numpy_datetime64("2022-04-15T21:10:28.360095+02:00"),),
            converted_currents,
        )[0],
        (223.3,),
    )

    # test several call and insure keep order
    numpy.testing.assert_almost_equal(
        deduce_machine_current(
            (
                str_datetime_to_numpy_datetime64("2022-01-15T21:07:58.360095+02:00"),
                str_datetime_to_numpy_datetime64("2022-04-15T21:10:28.360095+02:00"),
                str_datetime_to_numpy_datetime64("2022-04-15T21:08:58.360095+02:00"),
            ),
            converted_currents,
        ),
        (1.1, 223.3, 122.2),
    )


@pytest.mark.parametrize("n_part", (1, 3, 4, 7, 9, 12))
@pytest.mark.parametrize(
    "array_to_test",
    (numpy.arange(-15, 23, dtype=numpy.float32), numpy.ones(57, dtype=numpy.uint8)),
)
def test_split_timestamps(n_part, array_to_test):
    split_arrays = tuple(
        split_timestamps(
            my_array=array_to_test,
            n_part=n_part,
        )
    )
    assert len(split_arrays) == n_part
    assert split_arrays[0].dtype == array_to_test.dtype
    numpy.testing.assert_array_equal(
        numpy.concatenate([numpy.array(part) for part in split_arrays]),
        numpy.array(array_to_test),
    )


def test_get_entry_type(tmp_path):
    """test the get_entry_type function"""
    test_folder = tmp_path / "test_get_entry_type"
    test_folder.mkdir()
    test_file = os.path.join(test_folder, "test.hdf5")

    default_configuration = TomoHDF5Config()

    with HDF5File(test_file, mode="w") as h5f:
        h5f["group1/title"] = "darks"
        url_case1_darks = DataUrl(file_path=test_file, data_path="group1")

        h5f["group2/title"] = "flats"
        url_case1_flats = DataUrl(file_path=test_file, data_path="group2")

        h5f["group3/title"] = "darks"
        h5f["group3/technique/image_key"] = ImageKey.PROJECTION.value
        url_case2_proj = DataUrl(file_path=test_file, data_path="group3")

        h5f["group4/title"] = "darks"
        h5f["group4/technique/image_key"] = ImageKey.ALIGNMENT.value
        url_case2_alignment = DataUrl(file_path=test_file, data_path="group4")

        h5f["group5/technique/image_key"] = ImageKey.INVALID.value
        url_case3_invalid = DataUrl(file_path=test_file, data_path="group5")
        h5f["group6/technique/image_key"] = ImageKey.PROJECTION.value
        url_case3_proj = DataUrl(file_path=test_file, data_path="group6")

    # make sure if title is provided this is still working
    assert (
        get_entry_type(url_case1_darks, default_configuration) == AcquisitionStep.DARK
    )
    assert (
        get_entry_type(url_case1_flats, default_configuration) == AcquisitionStep.FLAT
    )

    # make sure the 'technique/image_key' get priority other the 'title'
    assert (
        get_entry_type(url_case2_proj, default_configuration)
        == AcquisitionStep.PROJECTION
    )
    assert (
        get_entry_type(url_case2_alignment, default_configuration)
        == AcquisitionStep.ALIGNMENT
    )

    # make sure if title is no more present then 'image_key' will be picked anyway
    assert get_entry_type(url_case3_invalid, default_configuration) is None
    assert (
        get_entry_type(url_case3_proj, default_configuration)
        == AcquisitionStep.PROJECTION
    )


def test_group_z_series(tmp_path):
    input_file = os.path.join(tmp_path, "test_z_series.h5")
    z_acquisitions = []
    n_series = 3
    n_scan_per_series = 4
    with h5py.File(input_file, mode="w") as h5f:
        i_scan = 0
        for i_series in range(n_series):
            for _ in range(n_scan_per_series):
                h5f[f"{i_scan}.1/sample/name"] = (
                    f"my_serie_{str(i_series).zfill(4)}_{str(i_scan).zfill(4)}"
                )
                z_acquisitions.append(
                    ZSeriesBaseAcquisition(
                        root_url=DataUrl(
                            file_path=input_file,
                            data_path=f"{i_scan}.1",
                            scheme="silx",
                        ),
                        configuration={},
                        detector_sel_callback=None,
                        start_index=i_scan,
                    )
                )
                i_scan += 1

    series_grouped = []
    for acquisition in z_acquisitions:
        series_grouped = group_series(
            acquisition=acquisition, list_of_series=series_grouped
        )

    assert len(series_grouped) == n_series
    for i in range(n_series):
        assert series_grouped[i] == list(
            [
                z_acquisitions[i * n_scan_per_series + i_scan]
                for i_scan in range(n_scan_per_series)
            ]
        )



# coding: utf-8

"""
module to convert fluo-tomo files (after PyMCA fit, tif files) to (nexus tomo compliant) .nx
"""

from __future__ import annotations

import pint
import logging
import os

from tqdm import tqdm

from .fluoscan import FluoTomoScan2D, FluoTomoScan3D

from nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill import utils
from nxtomomill.io.config.fluoconfig import TomoFluoConfig

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)

__all__ = [
    "from_fluo_to_nx",
]


def from_fluo_to_nx(
    configuration: TomoFluoConfig, progress: tqdm | None = None
) -> tuple:
    """
    Converts an fluo-tomo tiff files to a nexus file.
    For now duplicates data.

    :param configuration: configuration to use to process the data
    :param progress: if provided then will be updated with conversion progress
    :return: (nexus_file, entry)
    """
    if configuration.input_folder is None:
        raise ValueError("input_folder should be provided")
    if not os.path.isdir(configuration.input_folder):
        raise OSError(f"{configuration.input_folder} is not a valid folder path")

    if configuration.output_file is None:
        raise ValueError("output_file should be provided")

    if configuration.detectors is None:
        raise ValueError("Detector names should be provided.")

    fileout_h5 = utils.get_file_name(
        file_name=configuration.output_file,
        extension=configuration.file_extension,
        check=True,
    )

    if configuration.dimension == 2:
        scan = FluoTomoScan2D(
            scan=configuration.input_folder,
            dataset_basename=configuration.dataset_basename,
            detectors=configuration.detectors,
        )
    elif configuration.dimension == 3:
        scan = FluoTomoScan3D(
            scan=configuration.input_folder,
            dataset_basename=configuration.dataset_basename,
            detectors=configuration.detectors,
        )
    else:
        raise ValueError(f"Dimension should be 2 or 3 not {configuration.dimension}.")

    if progress is not None:
        progress.set_description("fluo2nx")
        progress.total = len(scan.el_lines)
        progress.n = 0

    _logger.info(f"Fluo lines preset in dataset are {scan.el_lines}")

    entry_list = []
    for element, lines in scan.el_lines.items():
        if progress is not None:
            progress.set_postfix_str(f"elmt - {element}")
            line_progress = tqdm(
                desc=f"elmt: {element} - line: ", position=1, leave=False
            )
            line_progress.total = len(lines)
        else:
            line_progress = None
        for i_line, line in enumerate(lines):
            if line_progress is not None:
                line_progress.set_postfix_str(f"elmt: {element} - line: {line}")

            for det in scan.detectors:
                elmt_line_data = scan.load_data(det, element=element, line_ind=i_line)
                if configuration.dimension == 2:
                    elmt_line_data = elmt_line_data.swapaxes(0, 1).copy()
                # Otherwise, it's 3D case, and the structure of elmt_line_data is OK.
                #
                my_nxtomo = NXtomo()
                my_nxtomo.instrument.detector.data = elmt_line_data
                my_nxtomo.instrument.detector.image_key_control = [
                    ImageKey.PROJECTION
                ] * elmt_line_data.shape[0]
                my_nxtomo.sample.rotation_angle = scan.rot_angles_deg * _ureg.degree
                my_nxtomo.instrument.detector.x_pixel_size = scan.pixel_size * _ureg(
                    "m"
                )
                my_nxtomo.instrument.detector.y_pixel_size = scan.pixel_size * _ureg(
                    "m"
                )

                # define a value to sample-detector and source-sample distance. To be set to the real value in the future
                my_nxtomo.instrument.detector.distance = 1.0 * _ureg.meter
                my_nxtomo.instrument.source.distance = 1.0 * _ureg.meter
                my_nxtomo.energy = scan.energy * _ureg.keV

                data_path = f"{det}_{element}_{line}"
                my_nxtomo.save(
                    file_path=fileout_h5,
                    data_path=data_path,
                    overwrite=configuration.overwrite,
                )
                entry_list.append((fileout_h5, data_path))
            if line_progress is not None:
                line_progress.update()
        if progress is not None:
            progress.update()
    return tuple(entry_list)
